__author__ = 'saeedamen'  # Saeed Amen

#
# Copyright 2016 Cuemacro
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and limitations under the License.
#
import pandas as pd
import numpy as np

from findatapy.market.datavendor import DataVendor
from findatapy.market.marketdatarequest import MarketDataRequest
from findatapy.util import ConfigManager
from findatapy.timeseries import Calculations

import abc
import datetime
import copy

class DataVendorBBG(DataVendor):
    """Abstract class for download of Bloomberg daily, intraday data and reference data.

    Implemented by:
        DataVendorBBGOpen - Adapted version of new Bloomberg Open API for Python which is recommended. Note that this
        requires compilation, via installed C++ compiler. For Python 3.5, this is Microsoft Visual Studio 2015.

        Or it is easier to install blpapi via conda, which deals with installing the DLL, adding the environment path etc.

        Note: no longer supports COM API, which is slower and only 32 bit

    """


    def __init__(self):
        super(DataVendorBBG, self).__init__()

    # implement method in abstract superclass
    def load_ticker(self, market_data_request):
        """Retrieves market data from external data source (in this case Bloomberg)

        Parameters
        ----------
        market_data_request : MarketDataRequest
            contains all the various parameters detailing time series start and finish, tickers etc

        Returns
        -------
        DataFrame
        """
        constants = DataConstants()

        market_data_request = MarketDataRequest(md_request=market_data_request)
        market_data_request_vendor = self.construct_vendor_market_data_request(market_data_request)

        data_frame = None

        logger = LoggerManager().getLogger(__name__)
        logger.info("Request Bloomberg data")

        # Do we need daily or intraday data?
        if (market_data_request.freq in ['daily', 'weekly', 'monthly', 'quarterly', 'yearly']):

            # Work out the fields which need to be downloaded via Bloomberg ref request (BDP) and
            # those that can be downloaded via Historical request (BDH)
            ref_fields = []
            ref_vendor_fields = []

            # Get user defined list of BBG fields/vendor fields which need to be downloaded by BDP
            bbg_ref_fields = list(constants.bbg_ref_fields.keys())
            bbg_ref_vendor_fields = list(constants.bbg_ref_fields.values())

            for i in range(0, len(market_data_request.fields)):
                if market_data_request.fields[i] in bbg_ref_fields \
                        or market_data_request_vendor.fields[i] in bbg_ref_vendor_fields:
                    ref_fields.append(market_data_request.fields[i])
                    ref_vendor_fields.append(market_data_request_vendor.fields[i])

            non_ref_fields = []
            non_ref_vendor_fields = []

            for i in range(0, len(market_data_request.fields)):
                if market_data_request.fields[i] not in bbg_ref_fields \
                        and market_data_request_vendor.fields[i] not in bbg_ref_vendor_fields:
                    non_ref_fields.append(market_data_request.fields[i])
                    non_ref_vendor_fields.append(market_data_request_vendor.fields[i])

            # For certain cases, need to use ReferenceDataRequest
            # eg. for events times/dates, last tradeable date fields (when specified)
            if len(ref_fields) > 0:

                # Careful: make sure you copy the market data request object (when threading, altering that can
                # cause concurrency issues!)
                old_fields = copy.deepcopy(market_data_request.fields)
                old_vendor_fields = copy.deepcopy(market_data_request_vendor.fields)

                # md_request = MarketDataRequest(md_request=market_data_request_copy)

                market_data_request.fields = ref_fields
                market_data_request.vendor_fields = ref_vendor_fields
                market_data_request_vendor = self.construct_vendor_market_data_request(market_data_request)

                # Just select those reference fields to download via reference
                datetime_data_frame = self.get_reference_data(market_data_request_vendor, market_data_request)

                # Download all the other event or non-ref fields (uses HistoricalDataRequest to Bloomberg)
                # concatenate with date time fields
                if len(non_ref_fields) > 0:

                    market_data_request.fields = non_ref_fields
                    market_data_request.vendor_fields = non_ref_vendor_fields
                    market_data_request_vendor = self.construct_vendor_market_data_request(market_data_request)

                    events_data_frame = self.get_daily_data(market_data_request, market_data_request_vendor)

                    col = events_data_frame.index.name
                    events_data_frame = events_data_frame.reset_index(drop=False)

                    data_frame = pd.concat([events_data_frame, datetime_data_frame], axis=1)
                    temp = data_frame[col]
                    del data_frame[col]
                    data_frame.index = temp
                else:
                    data_frame = datetime_data_frame

                market_data_request.fields = copy.deepcopy(old_fields)
                market_data_request_vendor.fields = copy.deepcopy(old_vendor_fields)

            # For all other daily/monthly/quarter data, we can use HistoricalDataRequest to Bloomberg
            else:
                data_frame = self.get_daily_data(market_data_request, market_data_request_vendor)

                # if data_frame is not None:
                #     # Convert fields with release-dt to dates (special case!) and assume everything else numerical
                #     for c in data_frame.columns:
                #         try:
                #             if 'release-dt' in c:
                #                 data_frame[c] = (data_frame[c]).astype('int').astype(str).apply(
                #                         lambda x: pd.to_datetime(x, format='%Y%m%d'))
                #             else:
                #                 data_frame[c] = pd.to_numeric(data_frame[c])
                #         except:
                #             pass

        # Assume one ticker only for intraday data and use IntradayDataRequest to Bloomberg
        if (market_data_request.freq in ['tick', 'intraday', 'second', 'minute', 'hourly']):
            market_data_request_vendor.tickers = market_data_request_vendor.tickers[0]

            if market_data_request.freq in ['tick', 'second']:
                data_frame = self.download_tick(market_data_request_vendor)
            else:
                data_frame = self.download_intraday(market_data_request_vendor)

            if data_frame is not None:
                if data_frame.empty:
                    try:
                        logger.info("No tickers returned for: " + market_data_request_vendor.tickers)
                    except:
                        pass

                    return None

                cols = data_frame.columns.values

                import pytz

                try:
                    data_frame = data_frame.tz_localize(pytz.utc)
                except:
                    data_frame = data_frame.tz_convert(pytz.utc)

                cols = market_data_request.tickers[0] + "." + cols
                data_frame.columns = cols

        logger.info("Completed request from Bloomberg.")

        return data_frame

    def get_daily_data(self, market_data_request, market_data_request_vendor):
        logger = LoggerManager().getLogger(__name__)

        data_frame = self.download_daily(market_data_request_vendor)

        # Convert from vendor to findatapy tickers/fields
        if data_frame is not None:
            if data_frame.empty:
                logger.info("No tickers returned for...")

                try:
                    logger.info(str(market_data_request_vendor.tickers))
                except:
                    pass

                return None

            returned_fields = data_frame.columns.get_level_values(0)
            returned_tickers = data_frame.columns.get_level_values(1)

            # TODO if empty try downloading again a year later
            try:
                fields = self.translate_from_vendor_field(returned_fields, market_data_request)
            except:
                print('Problem translating vendor field')

            tickers = self.translate_from_vendor_ticker(returned_tickers, market_data_request)

            ticker_combined = []

            for i in range(0, len(fields)):
                ticker_combined.append(tickers[i] + "." + fields[i])

            # Convert numerical columns to floats and dates to dates (avoids having object columns
            # which can cause issues with later Pandas)
            data_frame = self.force_type_conversion(data_frame)

            data_frame.columns = ticker_combined
            data_frame.index.name = 'Date'

            # Force sorting of index
            try:
                data_frame = data_frame.sort_index()
            except:
                pass

        return data_frame

    def get_reference_data(self, market_data_request_vendor, market_data_request):
        logger = LoggerManager().getLogger(__name__)

        constants = DataConstants()

        end = datetime.utcnow()

        from datetime import timedelta
        end = end + timedelta(days=365)  # because very often we may with to download data about future calendar events
        #  end.replace(year = end.year + 1)

        market_data_request_vendor.finish_date = end

        logger.debug("Requesting ref for " + market_data_request_vendor.tickers[0] + " etc.")

        data_frame = self.download_ref(market_data_request_vendor)

        logger.debug("Waiting for ref...")

        # Convert from vendor to findatapy tickers/fields
        if data_frame is not None:
            if data_frame.empty:
                return None

            returned_fields = data_frame.columns.get_level_values(0)
            returned_tickers = data_frame.columns.get_level_values(1)

        if data_frame is not None:
            # TODO if empty try downloading again a year later
            fields = self.translate_from_vendor_field(returned_fields, market_data_request)
            tickers = self.translate_from_vendor_ticker(returned_tickers, market_data_request)

            ticker_combined = []

            for i in range(0, len(fields)):
                ticker_combined.append(tickers[i] + "." + fields[i])

            data_frame.columns = ticker_combined

            # Need to convert numerical and datetime columns separately post pandas 0.23
            data_frame = self.force_type_conversion(data_frame)

            # data_frame = data_frame.apply(pd.to_datetime, errors='ignore')
            # data_frame = data_frame.apply(pd.to_numeric, errors='ignore')

            # TODO coerce will be deprecated from pandas 0.23.0 onwards) so remove!
            # data_frame = data_frame.convert_objects(convert_dates = 'coerce', convert_numeric= 'coerce')

        return data_frame


    def force_type_conversion(self, data_frame):
        constants = DataConstants()

        logger = LoggerManager().getLogger(__name__)

        if data_frame is not None:
            if not(data_frame.empty):
                # Need to convert numerical and datetime columns separately post pandas 0.23
                for c in data_frame.columns:
                    is_date = False

                    # Special case for ECO_RELEASE_DT / FIRST_REVISION_DATE
                    if 'ECO_RELEASE_DT' in c or 'FIRST_REVISION_DATE' in c:
                        try:
                            temp_col = []# data_frame[c].values
                            
                            for i in range(0, len(data_frame[c].values)):
                                try:
                                    temp_col.append(pd.to_datetime(str(int(data_frame[c].values[i])), format='%Y%m%d'))
                                except:
                                    temp_col.append(np.datetime64('NaT'))

                            data_frame[c] = temp_col
                        except Exception as e:
                            logger.warning("Couldn't convert " + str(c) + " to date.. was this column empty? " + str(e))

                    else:
                        # Only convert those Bloomberg reference fields to dates which have been listed explicitly
                        for d in constants.always_date_columns:
                            if d in c:
                                try:
                                    data_frame[c] = pd.to_datetime(data_frame[c], errors='coerce')

                                    is_date = True
                                    break
                                except:
                                    pass

                        # Otherwise this is not a date field so attempt to convert into numbers
                        if not(is_date):
                            try:
                                data_frame[c] = pd.to_numeric(data_frame[c], errors='ignore')
                            except:
                                pass

        logger.debug("Returning converted dataframe...")

        return data_frame

    # implement method in abstract superclass
    @abc.abstractmethod
    def kill_session(self):
        return

    @abc.abstractmethod
    def download_tick(self, market_data_request):
        return

    @abc.abstractmethod
    def download_intraday(self, market_data_request):
        return

    @abc.abstractmethod
    def download_daily(self, market_data_request, market_data_request_vendor):
        return

    @abc.abstractmethod
    def download_ref(self, market_data_request):
        return


#######################################################################################################################

import abc
import copy
import collections
import datetime
import re

try:
    import blpapi  # obtainable from Bloomberg website
except:
    pass

from findatapy.util.dataconstants import DataConstants
from findatapy.market.datavendorbbg import DataVendorBBG

from collections import defaultdict

class DataVendorBBGOpen(DataVendorBBG):
    """Calls the Bloomberg Open API to download market data: daily, intraday and reference data (needs blpapi).

    """

    def __init__(self):
        super(DataVendorBBGOpen, self).__init__()

    def download_tick(self, market_data_request):
        # Bloomberg OpenAPI implementation
        low_level_loader = BBGLowLevelTick()

        # by default we download all available fields!
        data_frame = low_level_loader.load_time_series(market_data_request)

        # self.kill_session() # need to forcibly kill_session since can't always reopen

        return data_frame

    def download_intraday(self, market_data_request):
        # Bloomberg OpenAPI implementation
        low_level_loader = BBGLowLevelIntraday()

        # by default we download all available fields!
        data_frame = low_level_loader.load_time_series(market_data_request)

        # self.kill_session() # need to forcibly kill_session since can't always reopen

        return data_frame

    def download_daily(self, market_data_request):
        # Bloomberg Open API implementation
        low_level_loader = BBGLowLevelDaily()

        # By default we download all available fields!
        data_frame = low_level_loader.load_time_series(market_data_request)

        # self.kill_session() # need to forcibly kill_session since can't always reopen

        return data_frame

    def download_ref(self, market_data_request):
        # Bloomberg Open API implementation
        low_level_loader = BBGLowLevelRef()

        market_data_request_vendor_selective = copy.copy(market_data_request)

        # special case for future date releases
        # if 'release-date-time-full' in md_request.fields:
        #     market_data_request_vendor_selective.fields = ['ECO_FUTURE_RELEASE_DATE_LIST']
        #
        # if 'last-tradeable-day' in md_request.fields:
        #     market_data_request_vendor_selective.fields = ['LAST_TRADEABLE_DT']

        data_frame = low_level_loader.load_time_series(market_data_request_vendor_selective)

        # self.kill_session() # need to forcibly kill_session since can't always reopen

        return data_frame

    def kill_session(self):
        # TODO not really needed, because we automatically kill sessions
        BBGLowLevelDaily().kill_session(None)
        BBGLowLevelRef().kill_session(None)
        BBGLowLevelIntraday().kill_session(None)


########################################################################################################################
#### Lower level code to interact with Bloomberg Open API

class BBGLowLevelTemplate(object):  # in order that the init function works in child classes

    convert_override_fields = {'settlement-calendar-code': 'SETTLEMENT_CALENDAR_CODE'}

    _session = None

    def __init__(self):
        self.RESPONSE_ERROR = blpapi.Name("responseError")
        self.SESSION_TERMINATED = blpapi.Name("SessionTerminated")
        self.CATEGORY = blpapi.Name("category")
        self.MESSAGE = blpapi.Name("message")

        return

    def load_time_series(self, market_data_request):

        # if(BBGLowLevelTemplate._session is None):
        logger = LoggerManager().getLogger(__name__)

        session = self.start_bloomberg_session()
        # else:
        #    session = BBGLowLevelTemplate._session

        def download_data_frame(session, eventQueue, options, cid):
            if options.security is not None:
                self.send_bar_request(session, eventQueue, options, cid)

                logger.info("Waiting for data to be returned...")

                return self.event_loop(session)
            else:
                logger.warn("No ticker or field specified!")

                return None

        try:
            # if can't open the session, kill existing one
            # then try reopen (up to 5 times...)
            i = 0

            while i < 5:
                if session is not None:
                    if not session.openService("//blp/refdata"):
                        logger.info("Try reopening Bloomberg session... try " + str(i))
                        self.kill_session(session)  # need to forcibly kill_session since can't always reopen
                        session = self.start_bloomberg_session()

                        if session is not None:
                            if session.openService("//blp/refdata"): i = 6
                else:
                    logger.info("Try opening Bloomberg session... try " + str(i))
                    session = self.start_bloomberg_session()

                i = i + 1

            # Give error if still doesn't work after several tries..
            if not session.openService("//blp/refdata"):
                logger.error("Failed to open //blp/refdata")

                return

            logger.info("Creating request...")

            eventQueue = blpapi.EventQueue()
            # eventQueue = None

            # Create a request
            from blpapi import CorrelationId

            options = self.fill_options(market_data_request)

            # In some instances we might split the options if need to have different overrides
            if isinstance(options, list):
                data_frame_list = []

                for op in options:
                    cid = CorrelationId()
                    data_frame_list.append(download_data_frame(session, eventQueue, op, cid))

                data_frame = Calculations().join(data_frame_list)
            else:
                cid = CorrelationId()
                data_frame = download_data_frame(session, eventQueue, options, cid)
        finally:
            # stop the session (will fail if NoneType)
            try:
                session.stop()
            except:
                pass

        return data_frame

    def event_loop(self, session):
        not_done = True

        data_frame_slice = None

        data_frame_list = []
        data_frame_cols = []

        while not_done:
            # nextEvent() method can be called with timeout to let
            # the program catch Ctrl-C between arrivals of new events
            event = session.nextEvent()  # removed time out
            # event = eventQueue.nextEvent()

            # Bloomberg will send us responses in chunks
            if event.eventType() == blpapi.Event.PARTIAL_RESPONSE:
                # logger.info("Processing Bloomberg Partial Response")
                data_frame_slice = self.process_response_event(event)
            elif event.eventType() == blpapi.Event.RESPONSE:
                # logger.info("Processing Bloomberg Full Response")
                data_frame_slice = self.process_response_event(event)
                not_done = False
            else:
                for msg in event:
                    if event.eventType() == blpapi.Event.SESSION_STATUS:

                        if msg.messageType() == self.SESSION_TERMINATED:
                            not_done = False

            # Append DataFrame only if not empty
            if data_frame_slice is not None:
                data_frame_list.append(self.combine_slices(data_frame_cols, data_frame_slice))

                # Keep list of columns we've already found (occasionally BBG seems to return columns more than once?)
                # (this will fail for intraday time series)
                try:
                    data_frame_cols.append(list(data_frame_slice.columns.get_level_values(1).values)[0])
                except:
                    pass

        from findatapy.timeseries import Calculations

        if data_frame_cols == [] and data_frame_list != []:  # intraday case
            data_frame = pd.concat(data_frame_list)
        else:  # daily frequencies
            data_frame = Calculations().join(data_frame_list, how='outer')

        # make sure we do not have any duplicates in the time series
        if data_frame is not None:
            # if data_frame.empty == False:
            data_frame.drop_duplicates(keep='last')

        return data_frame

    # Process raw message returned by Bloomberg
    def process_response_event(self, event):
        data_frame_list = []

        logger = LoggerManager().getLogger(__name__)

        for msg in event:
            # generates a lot of output - so don't use unless for debugging purposes
            # logger.info(msg)

            if msg.hasElement(self.RESPONSE_ERROR):
                logger.error("REQUEST FAILED: " + str(msg.getElement(self.RESPONSE_ERROR)))
                continue

            data_frame_slice = self.process_message(msg)

            if (data_frame_slice is not None):
                data_frame_list.append(data_frame_slice)

        if data_frame_list == []:
            logger.warn("No elements for ticker.")
            return None
        else:
            return pd.concat(data_frame_list)

    def get_previous_trading_date(self):
        tradedOn = datetime.date.today()

        while True:
            try:
                tradedOn -= datetime.timedelta(days=1)
            except OverflowError:
                return None

            if tradedOn.weekday() not in [5, 6]:
                return tradedOn

    # Create a session for Bloomberg with appropriate server & port
    def start_bloomberg_session(self):

        constants = DataConstants()
        tries = 0

        session = None

        logger = LoggerManager().getLogger(__name__)

        # Try up to 5 times to start a session
        while (tries < 5):
            try:
                # fill SessionOptions
                sessionOptions = blpapi.SessionOptions()
                sessionOptions.setServerHost(constants.bbg_server)
                sessionOptions.setServerPort(constants.bbg_server_port)

                logger.info("Starting Bloomberg session...")

                # create a Session
                session = blpapi.Session(sessionOptions)

                # start a Session
                if not session.start():
                    logger.error("Failed to start session.")
                    return

                logger.info("Returning session...")

                tries = 5
            except:
                tries = tries + 1

        # BBGLowLevelTemplate._session = session

        if session is None:
            logger.error("Failed to start session.")
            return

        return session

    def add_override(self, request, field, value):
        overrides = request.getElement("overrides")
        override1 = overrides.appendElement()
        override1.setElement("fieldId", field)
        override1.setElement("value", value)

    def add_override_dict(self, request, options):
        if options.overrides != {}:
            for k in options.overrides.keys():
                new_k = k

                # Is there a pretty name for this?
                if k in self.convert_override_fields:
                    new_k = self.convert_override_fields[k]

                self.add_override(request, new_k, options.overrides[k])

    @abc.abstractmethod
    def fill_options(self, market_data_request):
        pass

    @abc.abstractmethod
    def process_message(self, msg):
        # To be implemented by subclass
        return

    # Create request for data
    @abc.abstractmethod
    def send_bar_request(self, session, eventQueue, options, cid):
        # To be implemented by subclass
        return

    # create request for data
    @abc.abstractmethod
    def combine_slices(self, data_frame_cols, data_frame_slice):
        # To be implemented by subclass
        return

    def kill_session(self, session):
        logger = LoggerManager().getLogger(__name__)
        if (session is not None):
            try:
                session.stop()

                logger.info("Stopping session...")
            finally:
                logger.info("Finally stopping session...")

            session = None


class BBGLowLevelDaily(BBGLowLevelTemplate):

    def __init__(self):
        super(BBGLowLevelDaily, self).__init__()

    def combine_slices(self, data_frame_cols, data_frame_slice):
        # data
        try:
            if (data_frame_slice.columns.get_level_values(1).values[0]
                    not in data_frame_cols):
                # return data_frame.join(data_frame_slice, how="outer")
                return data_frame_slice
        except Exception as e:

            LoggerManager().getLogger(__name__).warn('Data slice empty ' + str(e))

            return None

        return None

    # Populate options for Bloomberg request for asset daily request
    def fill_options(self, market_data_request):
        constants = DataConstants()

        options = OptionsBBG()

        options.security = None # market_data_request.tickers
        options.startDateTime = market_data_request.start_date
        options.endDateTime = market_data_request.finish_date
        options.fields = market_data_request.fields

        options.overrides = market_data_request.overrides

        # options_copy = OptionsBBG(options_bbg=options)

        options_list = []

        last_overide_combo = {}

        if market_data_request.old_tickers is not None:

            ticker_list = []
            curr_options = OptionsBBG(options_bbg=options)

            for tick, old_tick in zip(market_data_request.tickers, market_data_request.old_tickers):
                if old_tick is not None:

                    t = old_tick.lower()

                    ## Special case for GDP where the advance, final and preliminary releases (but can define more
                    # in DataConstants)
                    # have the same ticker but different overrides
                    bbg_keyword_dict_override = constants.bbg_keyword_dict_override

                    # eg. RELEASE_STAGE_OVERRIDE
                    for bbg_override in bbg_keyword_dict_override.keys():

                        keyword_dict = bbg_keyword_dict_override[bbg_override]

                        for bbg_keyword in keyword_dict.keys():

                            # eg. ['gdp', 'advance']
                            keyword = keyword_dict[bbg_keyword]

                            if all(k.lower() in t for k in keyword):

                                if bbg_override in last_overide_combo:
                                    if last_overide_combo[bbg_override] != bbg_keyword:
                                        curr_options.security = ticker_list
                                        options_list.append(curr_options)

                                        ticker_list = []

                                        curr_options = OptionsBBG(options_bbg=options)

                                last_overide_combo[bbg_override] = bbg_keyword
                                curr_options.overrides[bbg_override] = bbg_keyword

                    ## Add other special cases
                    ticker_list.append(tick)

            if ticker_list != []:
                curr_options.security = ticker_list

            options_list.append(curr_options)
        else:
            options.security = market_data_request.tickers

            return options

        if len(options_list) == 1:
            return options_list[0]

        return options_list

    def process_message(self, msg):

        constants = DataConstants()
        # Process received events

        # SLOW loop (careful, not all the fields will be returned every time hence need to include the field name in the tuple)
        # perhaps try to run in parallel?
        logger = LoggerManager().getLogger(__name__)

        implementation = 'simple'

        if implementation == 'simple':
            ticker = msg.getElement('securityData').getElement('security').getValue()
            fieldData = msg.getElement('securityData').getElement('fieldData')

            data = defaultdict(dict)
            #
            # # FASTER avoid calling getValue/getElement methods in blpapi, very slow, better to cache variables
            for i in range(fieldData.numValues()):
                mini_field_data = fieldData.getValue(i)
                date = mini_field_data.getElement(0).getValue()

                for j in range(1, mini_field_data.numElements()):
                    field_value = mini_field_data.getElement(j)

                    data[(str(field_value.name()), ticker)][date] = field_value.getValue()

            # ORIGINAL repeated calling getValue/getElement much slower
            # for i in range(fieldData.numValues()):
            #     for j in range(1, fieldData.getValue(i).numElements()):
            #         data[(str(fieldData.getValue(i).getElement(j).name()), ticker)][fieldData.getValue(i).getElement(0).getValue()] \
            #             = fieldData.getValue(i).getElement(j).getValue()
        elif implementation == 'py4j':
            pass

            # TODO Py4J
            # from findatapy.market.bbgloop import bbgloop
            # from py4j.java_gateway import JavaGateway

            # gateway = JavaGateway()
            # data = gateway.entry_point.parseFieldDataArray(msg)
        elif implementation == 'cython':
            ticker = msg.getElement('securityData').getElement('security').getValue()
            fieldData = msg.getElement('securityData').getElement('fieldData')

            from findatapy.market.bbgloop import bbgloop

            data = bbgloop(fieldData, ticker)
        elif implementation == 'numba':
            ticker = msg.getElement('securityData').getElement('security').getValue()
            fieldData = msg.getElement('securityData').getElement('fieldData')

            from findatapy.market.bbgloop_numba import bbgloop_numba

            data = bbgloop_numba(fieldData, ticker)
            # TODO cython

        data_frame = pd.DataFrame(data)

        # If obsolete ticker could return no values
        if (not (data_frame.empty)):
            # data_frame.columns = pd.MultiIndex.from_tuples(data, names=['field', 'ticker'])
            data_frame.index = pd.to_datetime(data_frame.index)
            logger.info("Read: " + ticker + ' ' + str(data_frame.index[0]) + ' - ' + str(data_frame.index[-1]))
        else:
            return None

        return data_frame

    # Create request for data
    def send_bar_request(self, session, eventQueue, options, cid):
        logger = LoggerManager().getLogger(__name__)

        refDataService = session.getService("//blp/refdata")
        request = refDataService.createRequest("HistoricalDataRequest")

        request.set("startDate", options.startDateTime.strftime('%Y%m%d'))
        request.set("endDate", options.endDateTime.strftime('%Y%m%d'))

        # Only one security/eventType per request
        for field in options.fields:
            request.getElement("fields").appendValue(field)

        for security in options.security:
            request.getElement("securities").appendValue(security)

        # Add user defined overrides for BBG request
        self.add_override_dict(request, options)

        logger.info("Sending Bloomberg Daily Request:" + str(request))
        session.sendRequest(request=request, correlationId=cid)


class BBGLowLevelRef(BBGLowLevelTemplate):

    def __init__(self):
        super(BBGLowLevelRef, self).__init__()

    # Populate options for Bloomberg request for reference request
    def fill_options(self, market_data_request):
        options = OptionsBBG()

        options.security = market_data_request.tickers
        options.startDateTime = market_data_request.start_date
        options.endDateTime = market_data_request.finish_date
        options.fields = market_data_request.fields

        options.overrides = market_data_request.overrides

        return options

    def process_message(self, msg):
        logger = LoggerManager().getLogger(__name__)
        data = collections.defaultdict(dict)

        # process received events
        securityDataArray = msg.getElement('securityData')

        index = 0
        single = False

        for securityData in list(securityDataArray.values()):

            ticker = securityData.getElementAsString("security")
            fieldData = securityData.getElement("fieldData")

            for field in fieldData.elements():
                if not field.isValid():
                    field_name = "%s" % field.name()

                    logger.error(field_name + " is NULL")
                elif field.isArray():
                    # iterate over complex data returns.
                    field_name = "%s" % field.name()

                    for i, row in enumerate(field.values()):
                        try:
                            field_val = re.findall(r'"(.*?)"', "%s" % row)[0]
                        except:
                            e = row.getElement(0)
                            # k = str(e.name())
                            field_val = e.getValue()

                        data[(field_name, ticker)][index] = field_val

                        index = index + 1
                else:
                    field_name = "%s" % field.name()
                    data[(field_name, ticker)][0] = field.getValueAsString()

                    index = index + 1
                    single = True  # no need to create multi-index late, because just row!! CAREFUL!! needed for futures expiries

            fieldExceptionArray = securityData.getElement("fieldExceptions")

            for fieldException in list(fieldExceptionArray.values()):
                errorInfo = fieldException.getElement("errorInfo")

                print(errorInfo.getElementAsString("category"), ":", \
                      fieldException.getElementAsString("fieldId"))
                print("stop")

        # Explicitly state from_dict (buggy if create pd.DataFrame(data)
        data_frame = pd.DataFrame.from_dict(data)

        # Ff obsolete ticker could return no values
        if not(data_frame.empty):
            # if not(single):
            #    pass
            # data_frame.columns = pd.MultiIndex.from_tuples(data, names=['field', 'ticker'])

            logger.info("Reading: " + ticker + ' ' + str(data_frame.index[0]) + ' - ' + str(data_frame.index[-1]))
        else:
            return None

        return data_frame

    def combine_slices(self, data_frame_cols, data_frame_slice):
        if (data_frame_slice.columns.get_level_values(1).values[0] not in data_frame_cols):
            # return data_frame.join(data_frame_slice, how="outer")
            return data_frame_slice

        return None

    # Create request for data
    def send_bar_request(self, session, eventQueue, options, cid):
        logger = LoggerManager().getLogger(__name__)

        refDataService = session.getService("//blp/refdata")
        request = refDataService.createRequest('ReferenceDataRequest')

        self.add_override(request, 'TIME_ZONE_OVERRIDE', 23)  # force GMT time
        self.add_override(request, 'INCLUDE_EXPIRED_CONTRACTS', "Y")  # include expired contracts
        self.add_override(request, 'START_DT', options.startDateTime.strftime('%Y%m%d'))
        self.add_override(request, 'END_DT', options.endDateTime.strftime('%Y%m%d'))

        # Only one security/eventType per request
        for field in options.fields:
            request.getElement("fields").appendValue(field)

        for security in options.security:
            request.getElement("securities").appendValue(security)

        # Add user defined overrides for BBG request
        self.add_override_dict(request, options)

        logger.info("Sending Bloomberg Ref Request:" + str(request))
        session.sendRequest(request=request, correlationId=cid)


from operator import itemgetter


class BBGLowLevelIntraday(BBGLowLevelTemplate):

    def __init__(self):
        super(BBGLowLevelIntraday, self).__init__()

        # constants
        self.BAR_DATA = blpapi.Name("barData")
        self.BAR_TICK_DATA = blpapi.Name("barTickData")
        self.OPEN = blpapi.Name("open")
        self.HIGH = blpapi.Name("high")
        self.LOW = blpapi.Name("low")
        self.CLOSE = blpapi.Name("close")
        self.VOLUME = blpapi.Name("volume")
        self.NUM_EVENTS = blpapi.Name("numEvents")
        self.TIME = blpapi.Name("time")

    def combine_slices(self, data_frame_cols, data_frame_slice):
        # return data_frame.append(data_frame_slice)
        return data_frame_slice

    # Populate options for Bloomberg request for asset intraday request
    def fill_options(self, market_data_request):
        options = OptionsBBG()

        options.security = market_data_request.tickers[0]  # get 1st ticker only!
        options.event = market_data_request.trade_side.upper()
        options.barInterval = market_data_request.freq_mult
        options.startDateTime = market_data_request.start_date
        options.endDateTime = market_data_request.finish_date
        options.gapFillInitialBar = False
        options.overrides = market_data_request.overrides

        if hasattr(options.startDateTime, 'microsecond'):
            options.startDateTime = options.startDateTime.replace(microsecond=0)

        if hasattr(options.endDateTime, 'microsecond'):
            options.endDateTime = options.endDateTime.replace(microsecond=0)

        return options

    # iterate through Bloomberg output creating a DataFrame output
    # implements abstract method
    def process_message(self, msg):
        data = msg.getElement(self.BAR_DATA).getElement(self.BAR_TICK_DATA)

        logger = LoggerManager().getLogger(__name__)

        # logger.info("Processing intraday data for " + str(self._options.security))

        data_vals = list(data.values())

        # data_matrix = numpy.zeros([len(data_vals), 6])
        # data_matrix.fill(numpy.nan)
        #
        # date_index = [None] * len(data_vals)
        #
        # for i in range(0, len(data_vals)):
        #     data_matrix[i][0] = data_vals[i].getElementAsFloat(self.OPEN)
        #     data_matrix[i][1] = data_vals[i].getElementAsFloat(self.HIGH)
        #     data_matrix[i][2] = data_vals[i].getElementAsFloat(self.LOW)
        #     data_matrix[i][3] = data_vals[i].getElementAsFloat(self.CLOSE)
        #     data_matrix[i][4] = data_vals[i].getElementAsInteger(self.VOLUME)
        #     data_matrix[i][5] = data_vals[i].getElementAsInteger(self.NUM_EVENTS)
        #
        #     date_index[i] = data_vals[i].getElementAsDatetime(self.TIME)
        #
        # logger.info("Dates between " + str(date_index[0]) + " - " + str(date_index[-1]))
        #
        # # create pandas dataframe with the Bloomberg output
        # return pd.DataFrame(data = data_matrix, index = date_index,
        #                columns=['open', 'high', 'low', 'close', 'volume', 'events'])

        ## for loop method is touch slower
        # time_list = []
        # data_table = []

        # for bar in data_vals:
        #     data_table.append([bar.getElementAsFloat(self.OPEN),
        #                  bar.getElementAsFloat(self.HIGH),
        #                  bar.getElementAsFloat(self.LOW),
        #                  bar.getElementAsFloat(self.CLOSE),
        #                  bar.getElementAsInteger(self.VOLUME),
        #                  bar.getElementAsInteger(self.NUM_EVENTS)])
        #
        #     time_list.append(bar.getElementAsDatetime(self.TIME))

        # each price time point has multiple fields - marginally quicker
        tuple = [([bar.getElementAsFloat(self.OPEN),
                   bar.getElementAsFloat(self.HIGH),
                   bar.getElementAsFloat(self.LOW),
                   bar.getElementAsFloat(self.CLOSE),
                   bar.getElementAsInteger(self.VOLUME),
                   bar.getElementAsInteger(self.NUM_EVENTS)],
                  bar.getElementAsDatetime(self.TIME)) for bar in data_vals]

        data_table = list(map(itemgetter(0), tuple))
        time_list = list(map(itemgetter(1), tuple))

        try:
            logger.info("Dates between " + str(time_list[0]) + " - " + str(time_list[-1]))
        except:
            logger.info("No dates retrieved")
            return None

        # create pandas dataframe with the Bloomberg output
        return pd.DataFrame(data=data_table, index=time_list,
                                columns=['open', 'high', 'low', 'close', 'volume', 'events'])

    # Implement abstract method: create request for data
    def send_bar_request(self, session, eventQueue, options, cid):
        logger = LoggerManager().getLogger(__name__)
        refDataService = session.getService("//blp/refdata")
        request = refDataService.createRequest("IntradayBarRequest")

        # only one security/eventType per request
        request.set("security", options.security)
        request.set("eventType", options.event)
        request.set("interval", options.barInterval)

        # self.add_override(request, 'TIME_ZONE_OVERRIDE', 'GMT')

        if options.startDateTime is not None and options.endDateTime is not None:
            request.set("startDateTime", options.startDateTime)
            request.set("endDateTime", options.endDateTime)

        if options.gapFillInitialBar:
            request.append("gapFillInitialBar", True)

        # Add user defined overrides for BBG request
        self.add_override_dict(request, options)

        logger.info("Sending Intraday Bloomberg Request...")

        session.sendRequest(request=request, correlationId=cid)


class BBGLowLevelTick(BBGLowLevelTemplate):

    def __init__(self):
        super(BBGLowLevelTick, self).__init__()

        # constants
        self.TICK_DATA = blpapi.Name("tickData")
        self.COND_CODE = blpapi.Name("conditionCodes")
        self.TICK_SIZE = blpapi.Name("size")
        self.TIME = blpapi.Name("time")
        self.TYPE = blpapi.Name("type")
        self.VALUE = blpapi.Name("value")
        self.RESPONSE_ERROR = blpapi.Name("responseError")
        self.CATEGORY = blpapi.Name("category")
        self.MESSAGE = blpapi.Name("message")
        self.SESSION_TERMINATED = blpapi.Name("SessionTerminated")

    def combine_slices(self, data_frame, data_frame_slice):
        # return data_frame.append(data_frame_slice)
        return data_frame_slice

    # Populate options for Bloomberg request for asset intraday request
    def fill_options(self, market_data_request):
        options = OptionsBBG()

        options.security = market_data_request.tickers[0]  # Get 1st ticker only!
        options.event = market_data_request.trade_side.upper()
        # self._options.barInterval = md_request.freq_mult
        options.startDateTime = market_data_request.start_date
        options.endDateTime = market_data_request.finish_date
        # self._options.gapFillInitialBar = False

        if hasattr(options.startDateTime, 'microsecond'):
            options.startDateTime = options.startDateTime.replace(microsecond=0)

        if hasattr(options.endDateTime, 'microsecond'):
            options.endDateTime = options.endDateTime.replace(microsecond=0)

        return options

    # iterate through Bloomberg output creating a DataFrame output
    # implements abstract method
    def process_message(self, msg):
        data = msg.getElement(self.TICK_DATA).getElement(self.TICK_DATA)
        logger = LoggerManager().getLogger(__name__)

        #  logger.info("Processing tick data for " + str(self._options.security))

        data_vals = data.values()

        # for item in list(data_vals):
        #     if item.hasElement(self.COND_CODE):
        #         cc = item.getElementAsString(self.COND_CODE)
        #     else:
        #         cc = ""
        #
        #     # each price time point has multiple fields - marginally quicker
        #     tuple.append(([item.getElementAsFloat(self.VALUE),
        #                     item.getElementAsInteger(self.TICK_SIZE)],
        #                     item.getElementAsDatetime(self.TIME)))

        # slightly faster this way (note, we are skipping trade & CC fields)
        tuple = [([item.getElementAsFloat(self.VALUE),
                   item.getElementAsInteger(self.TICK_SIZE)],
                  item.getElementAsDatetime(self.TIME)) for item in data_vals]

        data_table = list(map(itemgetter(0), tuple))
        time_list = list(map(itemgetter(1), tuple))

        try:
            logger.info("Dates between " + str(time_list[0]) + " - " + str(time_list[-1]))
        except:
            logger.info("No dates retrieved")
            return None

        # create pandas dataframe with the Bloomberg output
        return pd.DataFrame(data=data_table, index=time_list,
                                columns=['close', 'ticksize'])

    # Implement abstract method: create request for data
    def send_bar_request(self, session, eventQueue, options, cid):
        logger = LoggerManager().getLogger(__name__)

        refDataService = session.getService("//blp/refdata")
        request = refDataService.createRequest("IntradayTickRequest")

        # only one security/eventType per request
        request.set("security", options.security)
        request.getElement("eventTypes").appendValue("TRADE")
        # request.set("eventTypes", self._options.event)
        request.set("includeConditionCodes", True)

        # self.add_override(request, 'TIME_ZONE_OVERRIDE', 'GMT')

        if options.startDateTime and options.endDateTime:
            request.set("startDateTime", options.startDateTime)
            request.set("endDateTime", options.endDateTime)

        # Add user defined overrides for BBG request
        self.add_override_dict(request, options)

        logger.info("Sending Tick Bloomberg Request...")

        session.sendRequest(request=request, correlationId=cid)


#######################################################################################################################

from findatapy.util.loggermanager import LoggerManager
from datetime import datetime


class OptionsBBG(object):

    def __init__(self, options_bbg=None, security=None, event=None, barInterval=None, startDateTime=None, endDateTime=None,
                 fields=None, overrides=None, gapFillInitialBar=False):

        if options_bbg is not None:
            security = copy.copy(options_bbg.security)
            event = copy.copy(options_bbg.event)
            barInterval = copy.copy(options_bbg.barInterval)
            startDateTime = copy.copy(options_bbg.startDateTime)
            endDateTime = copy.copy(options_bbg.endDateTime)
            fields = copy.copy(options_bbg.fields)
            overrides = copy.copy(options_bbg.overrides)
            gapFillInitialBar = copy.copy(options_bbg.gapFillInitialBar)

        self.security = security
        self.event = event
        self.barInterval = barInterval
        self.startDateTime = startDateTime
        self.endDateTime = endDateTime
        self.fields = fields
        self.overrides = overrides
        self.gapFillInitialBar = gapFillInitialBar
    
    @property
    def security(self):
        return self.__security

    @security.setter
    def security(self, security):
        self.__security = security
        
    @property
    def event(self):
        return self.__event

    @event.setter
    def event(self, event):
        self.__event = event
        
    @property
    def barInterval(self):
        return self.__barInterval

    @barInterval.setter
    def barInterval(self, barInterval):
        self.__barInterval = barInterval
        
    @property
    def startDateTime(self):
        return self.__startDateTime

    @startDateTime.setter
    def startDateTime(self, startDateTime):
        self.__startDateTime = startDateTime
        
    @property
    def endDateTime(self):
        return self.__endDateTime

    @endDateTime.setter
    def endDateTime(self, endDateTime):
        self.__endDateTime = endDateTime
    
    @property
    def fields(self):
        return self.__fields

    @fields.setter
    def fields(self, fields):
        self.__fields = fields
        
    @property
    def overrides(self):
        return self.__overrides

    @overrides.setter
    def overrides(self, overrides):
        self.__overrides = overrides
        
    @property
    def gapFillInitialBar(self):
        return self.__gapFillInitialBar

    @gapFillInitialBar.setter
    def gapFillInitialBar(self, gapFillInitialBar):
        self.__gapFillInitialBar = gapFillInitialBar


__author__ = 'saeedamen'  # Saeed Amen

#
# Copyright 2016-2021 Cuemacro
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and limitations under the License.
#


if __name__ == '__main__':
    ###### below line CRUCIAL when running Windows, otherwise multiprocessing doesn't work! (not necessary on Linux)
    from findatapy.util import SwimPool; SwimPool()

    from findatapy.market import Market, MarketDataRequest, MarketDataGenerator

    from findatapy.util import ConfigManager

    # findatapy has predefined tickers and categories in the conf folder in CSV files which you can modify, that
    # map from tickers/categories to vendor tickers/vendor fields
    #
    # time_series_categories_fields.csv
    # time_series_fields_list.csv
    # time_series_tickers_list.csv
    #
    # You can also add you own files with tickers, like we have done for FX vol like fx_vol_tickers.csv
    #
    # Having these ticker mappings makes it easier to mix and match data from different data sources, and saves us
    # having to remember vendor specific tickers

    cm = ConfigManager().get_instance()

    # Get all the categories for raw data (note this won't include generated categories like fx-vol-market,
    # which aggregate from many other categories)
    categories = list(cm.get_categories_from_tickers())

    print(categories)

    # Filter those categories which include quandl
    quandl_category = [x for x in categories if 'quandl' in x]

    print(quandl_category[0])

    # For this category, get all the tickers and fields which are available
    tickers = cm.get_tickers_list_for_category_str(categories[0])
    fields = cm.get_fields_list_for_category_str(categories[0])

    # We don't need to add the environment (eg. backtest)
    print("For category " + quandl_category[0] + ", tickers = " + str(tickers) + ", fields = " + str(fields))

    # Do a more complicated query, get any combinations
    df = cm.free_form_tickers_regex_query(category='fx/*|equities/*', ret_fields=['category', 'freq'])

    print(df.head(min(5, len(df.index))))

    # Let's search for all the category.data_source.freq.cut.tickers.fields combinations
    # where the category matches the regular expression for which has 'fx' or 'equities' in it and smart group
    df = cm.free_form_tickers_regex_query(category='fx/*|equities/*', ret_fields=['category', 'data_source', 'freq', 'cut', 'tickers', 'fields'],
                                          smart_group=True)

    print(df.head(min(5, len(df.index))))

    # We'll do the same query again, but this time, we'll convert it into a DataFrame which can use to create
    # a MarketDataRequest
    df_ungrouped = cm.free_form_tickers_regex_query(category='fx/*|equities/*', ret_fields=['category', 'data_source', 'freq', 'cut', 'tickers', 'fields'])

    print(df_ungrouped)

    # We can also do exact matches for multiple queries, where findatapy guesses the what parameters we were intending to find
    # so they don't necessarily have to be in the order category.data_source.freq.cut.ticker etc. They can be in whatever
    # order you want. Note that for this to work, we can not have values common across multiple fields, like tickers and category
    #
    # Here we are searching for the tickers/vendor_tickers combination which matches
    # - fx.daily.quandl (ie. findatapy will guess category='fx', freq='daily' and data_source = 'quandl')
    # - fx.daily.bloomberg (ie. findatapy
    df_guess = cm.free_form_tickers_query(['fx.daily.quandl', 'fx.daily.bloomberg'],
        ret_fields=['tickers', 'vendor_tickers'], list_query=True, smart_group=True)

    print(df_guess)

    market = Market(market_data_generator=MarketDataGenerator())

    md_request_list = market.create_md_request_from_dataframe(df_ungrouped)

    print(md_request_list)

    df_tickers = ConfigManager.get_dataframe_tickers()

    print(df_tickers)

    from findatapy.util import DataConstants

    constants = DataConstants()

    # You will likely need to put your own Quandl API key (free from Quandl's website!)
    quandl_api_key = constants.quandl_api_key
    print(market.fetch_market(md_request_str=categories[0], md_request=MarketDataRequest(quandl_api_key=quandl_api_key)))

    # Or we could create this download of FX Quandl data with a more free form query
    df_quandl_tickers = cm.free_form_tickers_regex_query(category='fx/*', data_source='quandl', cut='NYC')
    print(market.fetch_market(md_request_df=df_quandl_tickers, md_request=MarketDataRequest(quandl_api_key=quandl_api_key)))

    # Now lets get all Bloomberg data and Quandl FX data in our predefined tickers for the past week
    # we use ^fx$ as a regular expression, so it will only match with fx, and not any variations like fx-implied-vol
    md_request = MarketDataRequest(start_date='week', quandl_api_key=quandl_api_key)

    md_request_list = market.create_md_request_from_dataframe(cm.free_form_tickers_regex_query(category='^fx$', freq='daily', data_source='quandl|bloomberg', cut='NYC'),
                                                              md_request=md_request)

    for m in md_request_list:
        print(market.fetch_market(m))






__author__ = 'saeedamen'  # Saeed Amen

#
# Copyright 2016-2020 Cuemacro
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and limitations under the License.
#

# Note you'll need to "pip install s3fs" for this work which is not installed by default by findatapy
# You'll also need to have setup your S3 bucket, and have all your AWS credentials set on your machine
# This article explains how to give S3 rights to other/accounts-users https://stackoverflow.com/questions/45336781/amazon-s3-access-for-other-aws-accounts
# See below for an example

"""
{
    "Version": "2012-10-17",
    "Id": "S3AccessPolicy",
    "Statement": [
        {
            "Sid": "GiveFredAccess",
            "Effect": "Allow",
            "Principal": {
                "AWS": "arn:aws:iam::123456789012:user/fred"
            },
            "Action": [
                "s3:GetObject",
                "s3:ListBucket"
            ],
            "Resource": [
                "arn:aws:s3:::bucket-a",
                "arn:aws:s3:::bucket-a/*"
            ]
        }
    ]
}
"""

# It is recommended NOT to give your S3 access public in general

# NOTE: you need to make sure you have the correct data licences before storing data on disk (and whether other
# users can access it)

from findatapy.market import Market, MarketDataRequest

# In this case we are saving predefined tick tickers to disk, and then reading back
from findatapy.util.dataconstants import DataConstants
from findatapy.market.ioengine import IOEngine

# choose run_example = 0 for everything
# run_example = 1 - download FX tick data from Dukascopy and dump in S3
# run_example = 2 - list files in S3 according to a pattern

run_example = 2

folder = 's3://type_your_s3_bucket_here'

if run_example == 1 or run_example == 0:
    md_request = MarketDataRequest(
        start_date='04 Jan 2021',
        finish_date='05 Jan 2021',
        category='fx',
        data_source='dukascopy',
        freq='tick',
        tickers=['EURUSD'],
        fields=['bid', 'ask', 'bidv', 'askv'],
    )

    market = Market()

    df = market.fetch_market(md_request=md_request)

    print(df)

    # Save to disk in a format friendly for reading later (ie. s3://bla_bla_bla/backtest.fx.tick.dukascopy.NYC.EURUSD.parquet)
    # Here it will automatically generate the filename from the folder we gave
    # and the MarketDataRequest we made (altenatively, we could have just given the filename directly)
    IOEngine().write_time_series_cache_to_disk(folder, df, engine='parquet', md_request=md_request)

    md_request.data_engine = folder + '/*.parquet'

    df = market.fetch_market(md_request)

    print(df)

    # Or we could have just read it directly using
    df = IOEngine().read_time_series_cache_from_disk(folder, df, engine='parquet', md_request=md_request)

    # We can try this using daily data
    import os

    quandl_api_key = os.environ['QUANDL_API_KEY']

    df = market.fetch_market(md_request_str='fx.quandl.daily.NYC',
                             md_request=MarketDataRequest(start_date='01 Jan 2021', quandl_api_key=quandl_api_key))

    print(df)

if run_example == 2 or run_example == 0:

    io_engine = IOEngine()

    pattern = folder + "/*.parquet"

    matched_files = io_engine.list_files(pattern)

    print(matched_files)


from findatapy.market import Market, MarketDataRequest, MarketDataGenerator

market = Market(market_data_generator=MarketDataGenerator())

md_request = MarketDataRequest(start_date='07 Feb 2020', finish_date='20 Feb 2020', cut='LOC',
freq='daily', data_source='binance', category='crypto',
fields=['close', 'volume', 'quote-asset-volume'],
tickers=['XBTUSDT'])

df = market.fetch_market(md_request)
print(df.head(5))
print(df.tail(5))


__author__ = 'saeedamen'

from findatapy.timeseries.timezone import Timezone
from findatapy.timeseries.calendar import Calendar
from findatapy.timeseries.filter import Filter
from findatapy.timeseries.calculations import Calculations
from findatapy.timeseries.dataquality import DataQuality
from findatapy.timeseries.retstats import RetStats


__author__ = 'saeedamen' # Saeed Amen

#
# Copyright 2016 Cuemacro
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and limitations under the License.
#

"""
DataConstants

Has various constants required for the findatapy project. These have been defined as static variables.

"""

import os
import keyring


def path_join(folder, file):
    if 's3://' in folder:
        folder = folder.replace("s3://", "")
        folder = folder + "/" + file

        folder = folder.replace("//", "/")

        folder = "s3://" + folder

    else:
        if file[0] == '/':
            file = file[1::]

        folder = os.path.join(folder, file)

    folder = folder.replace("\\\\", "/")
    folder = folder.replace("\\", "/")

    return folder

def key_store(service_name):
    key = None

    # this will fail on some cloud notebook platforms so put in try/except loop
    try:
        key = keyring.get_password(service_name, os.getlogin())
    except:
        pass

    # set the keys by running set_api_keys.py file!

    # if key is None:
    #    key = input("Please enter the %s API key: " % service_name)
    #
    #    keyring.set_password(service_name, os.getlogin(), key)

    return key

class DataConstants(object):

    ###### SHOULD AUTODETECT FOLDER
    root_folder = os.path.dirname(os.path.dirname(os.path.abspath(__file__))).replace('\\', '/')
    temp_folder = root_folder + "temp"

    ###### FOR FUTURE VERSIONS (which include caching)
    # Folders for holding market data
    folder_historic_CSV = "x:/"
    folder_time_series_data = "x:/"

    # Usually the data folder where we want to store market data (eg. '.../test/*.parquet')
    # or 'arctic'
    default_data_engine = None

    ###### FOR DATABASE (Arctic/MongoDB)
    db_server = '127.0.0.1'
    db_port = '27017'
    db_username = None
    db_password = None

    ###### FOR TEMPORARY IN-MEMORY CACHE (Redis)
    db_cache_server = '127.0.0.1'
    db_cache_port = '6379'
    write_cache_engine = 'redis'  # 'redis' or 'no_cache' means we don't use cache

    use_cache_compression = True

    parquet_compression = 'gzip' # 'gzip' or 'snappy'

    # Note for AWS you can set these globally without having to specify here with AWS CLI
    cloud_credentials = {'aws_anon' : False}

    ## eg. {
    # {
    #    "aws_anon" : False,
    #    "aws_access_key": "asdfksesdf",
    #    "aws_secret_key": "asfsdf",
    #    "aws_access_token": "adsfsdf",
    # },

    ###### FOR ALIAS TICKERS
    # Config file for time series categories
    config_root_folder = path_join(root_folder, "conf")

    time_series_categories_fields = \
        path_join(config_root_folder, "time_series_categories_fields.csv")

    # We can have multiple tickers files (separated by ";")
    time_series_tickers_list = path_join(config_root_folder, "time_series_tickers_list.csv") +";" + \
                               path_join(config_root_folder, "fx_vol_tickers.csv")+";" + \
                               path_join(config_root_folder, "fx_forwards_tickers.csv")+";" + \
                               path_join(config_root_folder, "base_depos_tickers_list.csv")+";"

    time_series_fields_list = path_join(config_root_folder, "time_series_fields_list.csv")

    # Config file for long term econ data
    all_econ_tickers = path_join(config_root_folder, "all_econ_tickers.csv")
    econ_country_codes = path_join(config_root_folder, "econ_country_codes.csv")
    econ_country_groups = path_join(config_root_folder, "econ_country_groups.csv")

    holidays_parquet_table = path_join(config_root_folder, "holidays_table.parquet")

    # For events filtering
    events_category = 'events'
    events_category_dt = 'events_dt'

    # Ignore these columns when doing smart grouping
    drop_cols_smart_tickers_grouping = ['level_0']

    ###### FOR CURRENT VERSION

    # which marketdatagenerator type to use?
    # note - marketdatagenerator currently implemented
    #        cachedmarketdatagenerator is only for proprietary version at present
    default_market_data_generator = "marketdatagenerator"

    # In Python threading does not offer true parallisation, but can be useful when downloading data, because
    # a lot of the time is spend waiting on data, multiprocessing library addresses this problem by spawning new Python
    # instances, but this has greater overhead (maybe more advisable when downloading very long time series)

    # "thread" or "multiprocessing" (experimental!) library to use when downloading data
    market_thread_technique = "thread"

    multiprocessing_library = 'multiprocess' # 'multiprocessing_on_dill' or 'multiprocess' or 'multiprocessing'

    # How many threads to use for loading external data (don't do too many on slow machines!)
    # also some data sources will complain if you start too many parallel threads to call data!
    # for some data providers might get better performance from 1 thread only!
    market_thread_no = {             'quandl'      : 4,
                                     'bloomberg'   : 4,
                                     'yahoo'       : 1, # yfinance already threads requests, so don't do it twice!
                                     'other'       : 4,
                                     'dukascopy'   : 8,
                                     'fxcm'        : 4}

    # Seconds for timeout
    timeout_downloader = {'dukascopy' : 120}

    # Dukascopy specific settings
    dukascopy_retries = 20
    dukascopy_mini_timeout_seconds = 10
    dukascopy_multithreading = True # Can get rejected connections when threading with Dukascopy
    dukascopy_try_time = 0 # Usually values of 0-1/8-1/4-1 are reasonable
    # smaller values => quicker retry, but don't want to poll server too much

    # We can override the thread count and drop back to single thread for certain market data downloads, as can have issues with
    # quite large daily datasets from Bloomberg (and other data vendors) when doing multi-threading, so can override and use
    # single threading on these (and also split into several chunks)
    #
    override_multi_threading_for_categories = []

    # These fields should always be converted to numbers (for every data vendor in MarketDataGenerator)
    always_numeric_column = ['close', 'open', 'high', 'low', 'tot']

    # These fields will be forcibly be converted to datetime64 (only for Bloomberg)
    always_date_columns = ['release-date-time-full', 'last-tradeable-day',
                          'futures-chain-last-trade-dates', 'first-notice-date', 'first-tradeable-day',
                          'cal-non-settle-dates', 'first-revision-date', 'release-dt']

    default_time_units = 'us' # 'ns' or 'ms' too

    # These are string/object fields which do not need to be converted
    always_str_fields = ['futures-chain-tickers']

    # Dataframe chunk size
    chunk_size_mb = 500

    # Log config file
    logging_conf = path_join(config_root_folder, "logging.conf")

    ####### Bloomberg settings
    bbg_server = "localhost"       # needs changing if you use Bloomberg Server API
    bbg_server_port = 8194

    # These fields are BDS style fields to be downloaded using Bloomberg's Reference Data interface
    # You may need to add to this list
    bbg_ref_fields = {'release-date-time-full' : 'ECO_FUTURE_RELEASE_DATE_LIST',
                          'last-tradeable-day' : 'LAST_TRADEABLE_DT',
                          'futures-chain-tickers' :  'FUT_CHAIN',
                          'futures-chain-last-trade-dates' :'FUT_CHAIN_LAST_TRADE_DATES',
                          'first-notice-date' : 'FUT_NOTICE_FIRST',
                          'first-tradeable-day' : 'FUT_FIRST_TRADE_DT',
                          'cal-non-settle-dates': 'CALENDAR_NON_SETTLEMENT_DATES'
    }

    # Depending on the ticker field inclusion of specific keywords, apply a particular BBG override (make sure all lowercase)
    bbg_keyword_dict_override = {
        'RELEASE_STAGE_OVERRIDE' : {'A' : ['gdp', 'advance'],
                                    'F' : ['gdp', 'final'],
                                    'P' : ['gdp', 'preliminary']}
    }

    #######  Dukascopy settings
    dukascopy_base_url = "https://www.dukascopy.com/datafeed/"
    dukascopy_write_temp_tick_disk = False

    #######  FXCM settings
    fxcm_base_url = 'https://tickdata.fxcorporate.com/'
    fxcm_write_temp_tick_disk = False

    #######  Quandl settings
    quandl_api_key = key_store("Quandl")

    #######  Alpha Vantage settings
    alpha_vantage_api_key = key_store("AlphaVantage")

    #######  FXCM API (contact FXCM to get this)
    fxcm_api_key = "x"

    #######  Eikon settings
    eikon_api_key = key_store("Eikon")

    #######  Twitter settings (you need to set these up on Twitter)
    TWITTER_APP_KEY             = key_store("Twitter App Key")
    TWITTER_APP_SECRET          = key_store("Twitter App Secret")
    TWITTER_OAUTH_TOKEN	        = key_store("Twitter OAUTH token")
    TWITTER_OAUTH_TOKEN_SECRET	= key_store("Twitter OAUTH token Secret")

    ####### FRED (Federal Reserve of St Louis data) settings
    fred_api_key = key_store("FRED")

    ####### FX vol fields
    # Default download for FX vol surfaces etc.
    # types of quotation on vol surface
    # ATM, 25d riskies, 10d riskies, 25d strangles/butterflies, 10d strangles/butterflies
    fx_vol_part = ["V", "25R", "10R", "25B", "10B"]

    # Deltas quoted, eg 10d and 25d
    fx_vol_delta = [10, 25]

    # All the tenors on our vol surface
    fx_vol_tenor = ["ON", "1W", "2W", "3W", "1M", "2M", "3M", "4M", "6M", "9M", "1Y", "2Y", "3Y", "5Y"]

    # Which base depo currencies are available?
    base_depos_currencies = ['EUR', 'GBP', 'AUD', 'NZD', 'USD', 'CAD', 'CHF', 'NOK', 'SEK', 'JPY']

    # Tenors available for base depos
    base_depos_tenor = ["ON", "TN", "SN", "1W", "2W", "3W", "1M", "2M", "3M", "4M", "6M", "9M", "1Y", "2Y", "3Y", "5Y"]

    ### FX forwards total return index construction
    # All the tenors on our forwards
    fx_forwards_tenor = ["ON", "TN", "SN", "1W", "2W", "3W", "1M", "2M", "3M", "4M", "6M", "9M", "1Y", "2Y", "3Y", "5Y"]

    override_fields = {}

    ### What data environments are there
    default_data_environment = 'backtest'
    possible_data_environment = ['backtest', 'prod']

    # Overwrite field variables with those listed in DataCred or user provided dictionary override_fields
    def __init__(self, override_fields={}):
        try:
            from findatapy.util.datacred import DataCred
            cred_keys = DataCred.__dict__.keys()

            for k in DataConstants.__dict__.keys():
                if k in cred_keys and '__' not in k:
                    setattr(DataConstants, k, getattr(DataCred, k))
        except:
            pass

        # Store overrided fields
        if override_fields == {}:
            override_fields = DataConstants.override_fields
        else:
            DataConstants.override_fields = override_fields

        for k in override_fields.keys():
            if '__' not in k:
                setattr(DataConstants, k, override_fields[k])

    @staticmethod
    def reset_api_key(service_name, api_key):
        keyring.set_password(service_name, os.getlogin(), api_key)



__author__ = 'saeedamen' # Saeed Amen

#
# Copyright 2016 Cuemacro
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and limitations under the License.
#

import os

import csv
import pandas as pd


from findatapy.timeseries import Calculations
from findatapy.util.dataconstants import DataConstants
from findatapy.util.singleton import Singleton
from findatapy.util.loggermanager import LoggerManager

from dateutil.parser import parse

import re

import threading

class ConfigManager(object):
    """Functions for converting between vendor tickers and findatapy tickers (and vice-versa).

    """
    __metaclass__ = Singleton

    # Tickers and fields
    _dict_time_series_tickers_list_library_to_vendor = {}
    _dict_time_series_tickers_list_vendor_to_library = {}
    _dict_time_series_fields_list_vendor_to_library = {}
    _dict_time_series_fields_list_library_to_vendor = {}

    # Store expiry date
    _dict_time_series_ticker_expiry_date_library_to_library = {}

    # Store categories -> fields
    _dict_time_series_category_fields_library_to_library = {}
    _dict_time_series_category_startdate_library_to_library = {}
    _dict_time_series_category_revision_periods_library_to_library = {}
    _dict_time_series_category_tickers_library_to_library = {}

    # category, source, freq, ticker, cut, fields, sourceticker, local-close, expiry, ldn_clo
    _data_frame_time_series_tickers = None

    # Store categories ->
    _dict_time_series_tickers_list_library = {}

    __lock = threading.Lock()

    __instance = None

    def __init__(self, *args, **kwargs):
        pass

    def get_instance(cls, data_constants=None):
        if not ConfigManager.__instance:
            with ConfigManager.__lock:
                if not ConfigManager.__instance:
                    ConfigManager.__instance = super(ConfigManager, cls).__new__(ConfigManager)

                    if data_constants is None:
                        data_constants = DataConstants()

                    ConfigManager.__instance.populate_time_series_dictionaries(data_constants=data_constants)

        return ConfigManager.__instance


    ### time series ticker manipulators
    @staticmethod
    def populate_time_series_dictionaries(data_constants=None):

        if data_constants is None:
            data_constants = DataConstants()

        # There are several CSV files which contain data on the tickers

        # time_series_tickers_list - contains every tickers (findatapy tickers => vendor tickers)
        # category,	data_source,	freq, tickers, cut, fields, vendor_tickers (from your data provider)
        # eg. fx / bloomberg / daily / EURUSD / TOK / close,open,high,low / EURUSD CMPT Curncy

        # time_series_fields_list - translate findatapy fields name to vendor fields names
        # findatapy fields => vendor fields
        # data_source, fields, vendor_fields

        # time_series_categories_fields - for each category specific generic properties
        # category,	freq, data_source, fields, startdate
        # eg. fx / daily / bloomberg / close,high,low,open / 01-Jan-70

        # eg. bloomberg / close / PX_LAST

        ## Populate tickers list (allow for multiple files)

        if isinstance(data_constants.time_series_tickers_list, str):
            time_series_tickers_list_file = data_constants.time_series_tickers_list.split(';')
        else:
            time_series_tickers_list_file = data_constants.time_series_tickers_list

        df_tickers = []

        for tickers_list_file in time_series_tickers_list_file:

            if os.path.isfile(tickers_list_file):
                # reader = csv.DictReader(open(tickers_list_file))
                df = pd.read_csv(tickers_list_file)
                df = df.dropna(how='all')
                df_tickers.append(df)

                for index, line in df.iterrows():
                    category = line["category"]
                    data_source = line["data_source"]
                    freq_list = line["freq"].split(',')

                    if isinstance(freq_list, str):
                        freq_list = [freq_list]

                    for freq in freq_list:
                        tickers = line["tickers"]
                        cut = line["cut"]
                        vendor_tickers = line["vendor_tickers"]
                        expiry = None

                        try:
                            expiry = line['expiry']
                        except:
                            pass

                        if category != "":
                            # print("stop" + category + '.' +
                            #                                                  data_source + '.' +
                            #                                                  freq + '.' +
                            #                                                  cut + '.' +
                            #                                                  tickers)

                            # Conversion from library tickers to vendor vendor_tickers
                            ConfigManager._dict_time_series_tickers_list_library_to_vendor[category + '.' +
                                                                                 data_source + '.' +
                                                                                 freq + '.' +
                                                                                 cut + '.' +
                                                                                 tickers] = vendor_tickers

                            try:
                                if expiry != '':
                                    expiry = parse(expiry)
                                else:
                                    expiry = None
                            except:
                                pass

                            # Library of tickers by category
                            key = category + '.' + data_source + '.' + freq + '.' + cut

                            # Conversion from library tickers to library expiry date
                            ConfigManager._dict_time_series_ticker_expiry_date_library_to_library[
                                                                                           data_source + '.' +
                                                                                           tickers] = expiry

                            # Conversion from vendor vendor_tickers to library tickers
                            try:
                                ConfigManager._dict_time_series_tickers_list_vendor_to_library[
                                    key + '.' + vendor_tickers] = tickers
                            except:
                                logger.warning("Ticker not specified correctly (is some of this missing?) " + str(key) + '.' + str(vendor_tickers))

                            if key in ConfigManager._dict_time_series_category_tickers_library_to_library:
                                ConfigManager._dict_time_series_category_tickers_library_to_library[key].append(tickers)
                            else:
                                ConfigManager._dict_time_series_category_tickers_library_to_library[key] = [tickers]

        try:
            df_tickers = pd.concat(df_tickers).sort_values(by=['category', 'data_source', 'freq', 'cut'])
        except:
            pass

        try:
            df_tickers = df_tickers.reset_index()
        except:
            pass

        try:
            df_tickers = df_tickers.drop('level_0', axis=1).reset_index()
        except:
            pass

        ConfigManager._data_frame_time_series_tickers = df_tickers

        ## Populate fields conversions
        # reader = csv.DictReader(open(data_constants.time_series_fields_list))
        df = pd.read_csv(data_constants.time_series_fields_list)
        df = df.dropna(how='all')

        for index, line in df.iterrows():
            data_source = line["data_source"]
            fields = line["fields"]
            vendor_fields = line["vendor_fields"]

            # Conversion from vendor vendor_fields to library fields
            ConfigManager._dict_time_series_fields_list_vendor_to_library[data_source + '.' + vendor_fields] = fields

            # Conversion from library tickers to vendor vendor_fields
            ConfigManager._dict_time_series_fields_list_library_to_vendor[data_source + '.' + fields] = vendor_fields

        ## Populate categories fields list
        # reader = csv.DictReader(open(data_constants.time_series_categories_fields))
        df = pd.read_csv(data_constants.time_series_categories_fields)
        df = df.dropna(how='all')

        for index, line in df.iterrows():
            category = line["category"]
            data_source = line["data_source"]
            freq = line["freq"]
            cut = line["cut"]
            fields = line["fields"].split(',') # Can have multiple fields
            startdate = line["startdate"]
            revision_periods = line["revision_periods"]

            if category != "":
                # Conversion from library category to library fields list
                ConfigManager._dict_time_series_category_fields_library_to_library[
                    category + '.' + data_source + '.' + freq + '.' + cut] = fields

                # Conversion from library category to library startdate
                ConfigManager._dict_time_series_category_startdate_library_to_library[
                        category + '.' + data_source + '.' + freq + '.' + cut] = parse(startdate).date()

                # Conversion from library category to library revision periods
                ConfigManager._dict_time_series_category_revision_periods_library_to_library[
                    category + '.' + data_source + '.' + freq + '.' + cut] = revision_periods

    def free_form_tickers_regex_query(self, category=None, data_source=None, freq=None, cut=None, tickers=None, dict_filter={},
                                      ret_fields=['category', 'data_source', 'freq', 'cut'], smart_group=False):

        df = ConfigManager._data_frame_time_series_tickers

        if category is not None and not(df.empty):
            df = df[df['category'].str.match(category) == True]

        if data_source is not None and not(df.empty):
            df = df[df['data_source'].str.match(data_source) == True]

        if freq is not None and not(df.empty):
            df = df[df['freq'].str.match(freq) == True]

        if cut is not None and not(df.empty):
            df = df[df['cut'].str.match(cut) == True]

        if tickers is not None and not(df.empty):
            df = df[df['tickers'].str.match(tickers) == True]

        if cut is not None and not(df.empty):
            df = df[df['cut'].str.match(cut) == True]

        for k in dict_filter.keys():
            if k is not None and not(df.empty):
                df = df[df[k].str.match(dict_filter[k]) == True]

        if ret_fields is not None and not(df.empty):
            df = df[ret_fields]

        df = df.drop_duplicates()

        # Group any tickers/vendor_tickers together
        if smart_group:
            df = ConfigManager.smart_group_dataframe_tickers(df, ret_fields=ret_fields)

        return df

    def free_form_tickers_query(self, free_form_query, best_match_only=False, list_query=False,
                                ret_fields=['category', 'data_source', 'freq', 'cut', 'tickers', 'vendor_tickers', 'fields'],
                                smart_group=True):
        """From a string or list of properties for predefined tickers, we create a DataFrame that can be used to populate a
        MarketDataRequest. We search through all the predefined tickers, and "guess" any matches to our query, without
        having to use the standard query format which consists of category.data_source.freq.cut.ticker such as this example
        fx.bloomberg.daily.NYC.EURUSD.close

        eg. quandl.fx will match all tickers which are from "quandl" and have a "category" fx

        We must be careful to make sure that categories, data_sources  etc. are unique and do not overlap with other properties
        like tickers

        Parameters
        ----------
        free_form_query : str
            A query that can be used to generate a MarketDataRequest

            eg. quandl.fx

        best_match_only : bool
            Only return at most 1 row of a DataFrame (default: False)

        list_query : bool
            Is this a list of tickers?

        ret_fields : str(list)
            Which properties of a MarketDataRequest to return

        smart_group : bool
            Smart group tickers of a particular category in a specific row

        Returns
        -------
        DataFrame
        """
        logger = LoggerManager().getLogger(__name__)

        logger.info("Finding ticker combination which matches " + str(free_form_query))

        df = ConfigManager._data_frame_time_series_tickers

        if list_query and isinstance(free_form_query, list):
            free_form_query = free_form_query
        elif ',' in free_form_query:
            free_form_query = free_form_query.split(',')
        else:
            free_form_query = [free_form_query]

        df_joined_list = []

        for key in free_form_query:
            df_joined = df

            key = ConfigManager.split_ticker_string(key)

            # Search through all the keywords, and see if matches with any columns of our predefined tickers
            try:
                for k in key:
                    for c in df.columns:
                        try:
                            df_temp = df_joined[df_joined[c] == k]
                        except:
                            df_temp = pd.DataFrame()

                        if not(df_temp.empty):
                            df_joined = df_temp
                            break

                df_joined_list.append(df_joined)
            except Exception as e:
                pass

        # Drop any duplicated tickers
        df = pd.concat(df_joined_list).drop_duplicates()

        if len(df.index) > 1:
            logger.info("Found multiple matches for ticker combination, first trying smart group...")

            if smart_group:
                df = self.smart_group_dataframe_tickers(df, ret_fields=ret_fields)

            if best_match_only:
                logger.info("Taking only top match...")
                df = pd.DataFrame(df.head(1))

        if ret_fields is not None and not(df.empty):
            df = df[ret_fields]

        return df

    @staticmethod
    def split_ticker_string(md_request_str):

        if isinstance(md_request_str, str):
            split_lst = []

            word = ''
            ignore_dot = False

            for c in md_request_str:
                if c == '{':
                    ignore_dot = True

                elif c == '}':
                    ignore_dot = False

                elif c == '.' and not (ignore_dot):
                    split_lst.append(word)
                    word = ''
                    ignore_dot = False
                else:
                    word = word + c

            split_lst.append(word)

            return split_lst

        return md_request_str

    @staticmethod
    def smart_group_dataframe_tickers(df, ret_fields=['category', 'data_source', 'freq', 'cut']):
        """Groups together a DataFrame of metadata associated with assets, which can be used to create MarketDataRequest
        objects
        """

        if ret_fields is None:
            ret_fields = df.columns.to_list()
        elif isinstance(ret_fields, str):
            if ret_fields == 'all':
                ret_fields = df.columns.to_list()
        elif isinstance(ret_fields, list):
            if ret_fields == []:
                ret_fields =  df.columns.to_list()

        if set(['category', 'data_source', 'freq', 'cut']).issubset(ret_fields):
            group_fields = ret_fields.copy()

            agg_dict = {}

            if 'tickers' in ret_fields:
                df['tickers'] = [[x] for x in df['tickers'].tolist()]
                agg_dict['tickers'] = 'sum'
                group_fields.remove('tickers')

            if 'vendor_tickers' in ret_fields:
                df['vendor_tickers'] = [[x] for x in df['vendor_tickers'].tolist()]
                agg_dict['vendor_tickers'] = 'sum'
                group_fields.remove('vendor_tickers')

            if agg_dict != {}:

                try:
                    df = df.drop(data_constants.drop_cols_smart_tickers_grouping, axis=1)
                except:
                    pass

                df_temp = df.groupby(group_fields).agg(agg_dict)

                # If grouping fails (when there aren't multiple elements to group!)
                if df_temp.empty:
                   pass
                else:
                    for i, g in enumerate(group_fields):
                        df_temp[g] = df_temp.index.get_level_values(i)

                    df = df_temp.reset_index(drop=True)

        return df

    @staticmethod
    def get_dataframe_tickers():
        return ConfigManager._data_frame_time_series_tickers

    @staticmethod
    def get_categories_from_fields():
        return ConfigManager._dict_time_series_category_fields_library_to_library.keys()

    @staticmethod
    def get_categories_from_tickers():
        return ConfigManager._dict_time_series_category_tickers_library_to_library.keys()

    @staticmethod
    def get_categories_from_tickers_selective_filter(filter):
        initial_list = ConfigManager._dict_time_series_category_tickers_library_to_library.keys()
        filtered_list = []

        for category_desc in initial_list:
            split_cat = category_desc.split('.')

            category = split_cat[0]
            # data_source = split_cat[1]
            # freq = split_cat[2]
            # cut = split_cat[3]

            if filter in category:
                filtered_list.append(category_desc)

        return filtered_list

    @staticmethod
    def get_potential_caches_from_tickers():
        all_categories = ConfigManager._dict_time_series_category_tickers_library_to_library.keys()

        expanded_category_list = []

        for sing in all_categories:
            split_sing = sing.split(".")
            category = split_sing[0]
            data_source = split_sing[1]
            freq = split_sing[2]
            cut = split_sing[3]

            if(freq == 'intraday'):
                intraday_tickers = ConfigManager.get_tickers_list_for_category(category, data_source, freq, cut)

                for intraday in intraday_tickers:
                    expanded_category_list.append(category + '.' + data_source + '.' + freq +
                                                  '.' + cut + '.' + intraday)
            else:
                expanded_category_list.append(category + '.' + data_source + '.' + freq +
                                                  '.' + cut)

        return expanded_category_list

    @staticmethod
    def get_fields_list_for_category(category, data_source, freq, cut):
        return ConfigManager._dict_time_series_category_fields_library_to_library[
                category + '.' + data_source + '.' + freq + '.' + cut]

    @staticmethod
    def get_fields_list_for_category_str(category):
        return ConfigManager._dict_time_series_category_fields_library_to_library[category]

    @staticmethod
    def get_startdate_for_category(category, source, freq, cut):
        return ConfigManager._dict_time_series_category_startdate_library_to_library[
                category + '.' + source + '.' + freq + '.' + cut]

    @staticmethod
    def get_revision_periods_for_category(category, source, freq, cut):
        return ConfigManager._dict_time_series_category_revision_periods_library_to_library[
                category + '.' + source + '.' + freq + '.' + cut]

    @staticmethod
    def get_expiry_for_ticker(data_source, ticker):
        return ConfigManager._dict_time_series_ticker_expiry_date_library_to_library[
                data_source + '.' + ticker]

    @staticmethod
    def get_filtered_tickers_list_for_category(category, data_source, freq, cut, filter):
        tickers = ConfigManager._dict_time_series_category_tickers_library_to_library[
                category + '.' + data_source + '.' + freq + '.' + cut]

        filtered_tickers = []

        for tick in tickers:
            if re.search(filter, tick):
                filtered_tickers.append(tick)

        return filtered_tickers

    @staticmethod
    def get_tickers_list_for_category(category, data_source, freq, cut):
        return ConfigManager._dict_time_series_category_tickers_library_to_library[
                category + '.' + data_source + '.' + freq + '.' + cut]

    @staticmethod
    def get_vendor_tickers_list_for_category(category, data_source, freq, cut):
        category_source_freq_cut = category + '.' + data_source + '.' + freq + '.' + cut

        return ConfigManager.get_vendor_tickers_list_for_category_str(category_source_freq_cut)

    @staticmethod
    def get_tickers_list_for_category_str(category_data_source_freq_cut):
        return ConfigManager._dict_time_series_category_tickers_library_to_library[category_data_source_freq_cut]

    @staticmethod
    def get_vendor_tickers_list_for_category_str(category_data_source_freq_cut):
        tickers = ConfigManager._dict_time_series_category_tickers_library_to_library[category_data_source_freq_cut]

        vendor_tickers = []

        for t in tickers:
            vendor_tickers.append(ConfigManager.convert_library_to_vendor_ticker_str(category_data_source_freq_cut + "." + t))

        return ConfigManager.flatten_list_of_lists(vendor_tickers)

    @staticmethod
    def convert_library_to_vendor_ticker(category, data_source, freq, cut, ticker):
        return ConfigManager._dict_time_series_tickers_list_library_to_vendor[
            category + '.' + data_source + '.' + freq + '.' + cut + '.' + ticker]

    @staticmethod
    def convert_library_to_vendor_ticker_str(category_data_source_freq_cut_ticker):
        return ConfigManager._dict_time_series_tickers_list_library_to_vendor[category_data_source_freq_cut_ticker]

    @staticmethod
    def convert_vendor_to_library_ticker(category, data_source, freq, cut, vendor_tickers):
        return ConfigManager._dict_time_series_tickers_list_vendor_to_library[
            category + '.' + data_source + '.' + freq + '.' + cut + '.' + vendor_tickers]

    @staticmethod
    def convert_vendor_to_library_field(data_source, vendor_fields):
        return ConfigManager._dict_time_series_fields_list_vendor_to_library[
            data_source + '.' + vendor_fields]

    @staticmethod
    def convert_library_to_vendor_field(data_source, fields):
        return ConfigManager._dict_time_series_fields_list_library_to_vendor[
            data_source + '.' + fields]

    @staticmethod
    def remove_duplicates_and_flatten_list(lst):
        return list(dict.fromkeys(ConfigManager.flatten_list_of_lists(lst)))

    @staticmethod
    def flatten_list_of_lists(list_of_lists):
        """Flattens lists of obj, into a single list of strings (rather than characters, which is default behavior).

        Parameters
        ----------
        list_of_lists : obj (list)
            List to be flattened

        Returns
        -------
        str (list)
        """

        if isinstance(list_of_lists, list):
            rt = []
            for i in list_of_lists:
                if isinstance(i, list):
                    rt.extend(ConfigManager.flatten_list_of_lists(i))
                else:
                    rt.append(i)

            return rt

        return list_of_lists


## test function
if __name__ == '__main__':
    logger = LoggerManager().getLogger(__name__)

    data_constants = DataConstants(override_fields={'use_cache_compression' : False})

    print(data_constants.use_cache_compression)

    cm = ConfigManager().get_instance()

    categories = cm.get_categories_from_fields()

    logger.info("Categories from fields list")
    print(categories)

    categories = cm.get_categories_from_tickers()

    logger.info("Categories from tickers list")
    print(categories)

    filter = 'events'

    categories_filtered = cm.get_categories_from_tickers_selective_filter(filter)
    logger.info("Categories from tickers list, filtered by events")
    print(categories_filtered)

    logger.info("For each category, print all tickers and fields")

    for sing in categories:
        split_sing = sing.split(".")
        category = split_sing[0]
        data_source = split_sing[1]
        freq = split_sing[2]
        cut = split_sing[3]

        logger.info("tickers for " + sing)
        tickers = cm.get_tickers_list_for_category(category, data_source, freq, cut)

        print(tickers)

        logger.info("fields for " + sing)
        fields = cm.get_fields_list_for_category(category, data_source, freq, cut)

        print(fields)

    # test the various converter mechanisms
    output = cm.convert_library_to_vendor_ticker(category='fx', data_source='bloomberg', freq='daily', cut='TOK', ticker='USDJPY')

    print(output)

    output = cm.convert_vendor_to_library_ticker(
        data_source='bloomberg', vendor_tickers='EURUSD CMPT Curncy')

    print(output)

    output = cm.convert_vendor_to_library_field(
        data_source='bloomberg', vendor_fields='PX_LAST')

    print(output)

    output = cm.convert_library_to_vendor_field(
        data_source='bloomberg', vendor_fields='close')

    print(output)

    print(DataConstants().use_cache_compression)






__author__ = 'saeedamen'  # Saeed Amen

#
# Copyright 2016-2020 Cuemacro
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and limitations under the License.
#

if __name__ == '__main__':
    ###### below line CRUCIAL when running Windows, otherwise multiprocessing doesn't work! (not necessary on Linux)
    from findatapy.util import SwimPool; SwimPool()

    # choose run_example = 0 for everything
    # run_example = 1 - download EURUSD free tick data from DukasCopy example
    # run_example = 2 - download S&P500 free tick data from DukasCopy example

    run_example = 0

    if run_example == 1 or run_example == 0:

        ####### DukasCopy examples
        # let's download data for 14 Jun 2016 for EUR/USD - the raw data has bid/ask, if we specify close, we calculate
        # it as the average

        from findatapy.market import Market, MarketDataRequest, MarketDataGenerator

        market = Market(market_data_generator=MarketDataGenerator())

        # first we can do it by defining all the vendor fields, tickers etc. so we bypass the configuration file
        md_request = MarketDataRequest(start_date='14 Jun 2016', finish_date='20 Jun 2016',
                                       fields=['bid'], vendor_fields=['bid'],
                                       freq='tick', data_source='dukascopy',
                                       tickers=['EURUSD'], vendor_tickers=['EURUSD'])

        df = market.fetch_market(md_request)
        print(df.tail(n=10))

    if run_example == 2 or run_example == 0:
        ####### Dukascopy S&P500 example
        from findatapy.market import Market, MarketDataRequest, MarketDataGenerator

        market = Market(market_data_generator=MarketDataGenerator())

        md_request = MarketDataRequest(start_date='14 Jun 2016', finish_date='15 Jun 2016',
                                       fields=['bid', 'ask'], vendor_fields=['bid', 'ask'],
                                       freq='tick', data_source='dukascopy',
                                       tickers=['S&P500'], vendor_tickers=['USA500IDXUSD'])

        # Careful need to divide by 1000.0
        df = market.fetch_market(md_request) / 1000.0

        print(df)



__author__ = 'saeedamen' # Saeed Amen

#
# Copyright 2016 Cuemacro
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and limitations under the License.
#

import logging
import logging.config

from findatapy.util.dataconstants import DataConstants
from findatapy.util.singleton import Singleton

class LoggerManager(object):
    """Acts as a wrapper for logging.

    """
    __metaclass__ = Singleton

    _loggers = {}

    def __init__(self, *args, **kwargs):
        pass

    @staticmethod
    def getLogger(name=None):
        if not name:
            try:
                logging.config.fileConfig(DataConstants().logging_conf)
            except: pass

            log = logging.getLogger();
        elif name not in LoggerManager._loggers.keys():
            try:
                logging.config.fileConfig(DataConstants().logging_conf)
            except: pass

            LoggerManager._loggers[name] = logging.getLogger(str(name))

        log = LoggerManager._loggers[name]

        # when recalling appears to make other loggers disabled
        # hence apply this hack!
        for name in LoggerManager._loggers.keys():
            LoggerManager._loggers[name].disabled = False

        return log

if __name__ == '__main__':

    logger = LoggerManager.getLogger(__name__)

    logger.info("Hello")


__author__ = 'saeedamen'  # Saeed Amen

#
# Copyright 2016 Cuemacro
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and limitations under the License.
#

import copy
from findatapy.util import ConfigManager
from findatapy.util import DataConstants
from findatapy.market.ioengine import SpeedCache

import concurrent.futures

import json

constants = DataConstants()

# from deco import *

class Market(object):
    """Higher level class which fetches market data using underlying classes such as MarketDataGenerator.

    Also contains several other classes, which are for asset specific instances, for example for generating FX spot time series
    or FX volatility surfaces.
    """

    def __init__(self, market_data_generator=None, md_request=None):
        if market_data_generator is None:
            if constants.default_market_data_generator == "marketdatagenerator":
                from findatapy.market import MarketDataGenerator
                market_data_generator = MarketDataGenerator()
            elif constants.default_market_data_generator == 'cachedmarketdatagenerator':
                # NOT CURRENTLY IMPLEMENTED FOR FUTURE USE
                from finaddpy.market import CachedMarketDataGenerator
                market_data_generator = CachedMarketDataGenerator()
            else:
                from findatapy.market import MarketDataGenerator
                market_data_generator = MarketDataGenerator()

        self.speed_cache = SpeedCache()
        self._market_data_generator = market_data_generator
        self._filter = Filter()
        self._calculations = Calculations()
        self.md_request = md_request

    def fetch_market(self, md_request=None, md_request_df=None, md_request_str=None, tickers=None,
                     start_date=None, finish_date=None, best_match_only=False, **kwargs):
        """Fetches market data for specific tickers

        The user does not need to know to the low level API for each data provider works. The MarketDataRequest
        needs to supply parameters that define each data request. It has details which include:
            ticker eg. EURUSD
            field eg. close
            category eg. fx
            data_source eg. bloomberg
            start_date eg. 01 Jan 2015
            finish_date eg. 01 Jan 2017

        It can also have many optional attributes, such as
            vendor_ticker eg. EURUSD Curncy
            vendor_field eg. PX_LAST

        We can also create MarketDataRequest objects, using strings eg. fx.quandl.daily.NYC.EURUSD (category.data_source.freq.cut.ticker),
        as Python dict, as DataFrame with various properties (like ticker, category etc.)

        Parameters
        ----------
        md_request : MarketDataRequest (or shorthand str/DataFrame/dict)
            Describing what market data to fetch

        md_request_df : DataFrame
            Another way to specify some of the data request properties

        md_request_str : str
            We can specify part of the MarketDataRequest as a string, which gets converted later (or a JSON object)

        Returns
        -------
        pd.DataFrame
            Contains the requested market data

        """

        if self.md_request is not None:
            md_request = self.md_request

        # Any kwargs are assumed to be to set MarketDataRequest attributes
        if kwargs != {}:
            md_request = self._kwargs_to_md_request(kwargs, md_request)

        # When we have specified a string
        if md_request_str is not None:
            md_request = self.create_md_request_from_str(md_request_str, md_request=md_request,
                                                         start_date=start_date, finish_date=finish_date, best_match_only=best_match_only, **kwargs)

            return self.fetch_market(md_request)

        # When we have specified predefined tickers
        if tickers is not None:
            md_request = self.create_md_request_from_tickers(tickers, md_request=md_request,
                                                         start_date=start_date, finish_date=finish_date, best_match_only=best_match_only, **kwargs)

            return self.fetch_market(md_request)

        # When we have specified a DataFrame with tickers
        if md_request_df is not None:
            md_request = self.create_md_request_from_dataframe(md_request_df, md_request=md_request,
                                                               start_date=start_date, finish_date=finish_date, best_match_only=best_match_only, **kwargs)

            return self.fetch_market(md_request)

        # Or directly as a string
        if isinstance(md_request, str):
            md_request = self.create_md_request_from_str(md_request, start_date=start_date, finish_date=finish_date, best_match_only=best_match_only, **kwargs)

            return self.fetch_market(md_request)

        # Or directly as a dict
        if isinstance(md_request, dict):
            md_request = self.create_md_request_from_dict(md_request, start_date=start_date, finish_date=finish_date)

            return self.fetch_market(md_request)

        # Or directly as a DataFrame
        if isinstance(md_request, pd.DataFrame):
            md_request = self.create_md_request_from_dataframe(md_request, start_date=start_date, finish_date=finish_date, **kwargs)

            return self.fetch_market(md_request)

        data_frame = None

        if isinstance(md_request, list):
            md_request = self.flatten_list_of_lists(md_request)

            if len(md_request) == 1:
                md_request = md_request[0]

        # If we've got a list MarketDataRequest objects, use threading to independently call them
        if isinstance(md_request, list):
            if len(md_request) > 0:

                with concurrent.futures.ThreadPoolExecutor(max_workers=md_request[0].list_threads) as executor:
                    df_list = list(executor.map(self.fetch_market, md_request))

                df_filtered_list = []

                for md, df in zip(md_request, df_list):

                    columns = []

                    for tick in md.tickers:
                        for fiel in md.fields:
                            columns.append(tick + "." + fiel)

                    if df is not None:
                        if df.empty:
                            df = pd.DataFrame(columns=columns)
                    else:
                        df = pd.DataFrame(columns=columns)

                    df_filtered_list.append(df)

                return self._calculations.join(df_filtered_list)
            else:
                return None

        key = md_request.generate_key()

        # If internet_load has been specified don't bother going to cache (might end up calling lower level cache though
        # through MarketDataGenerator)
        if 'cache_algo' in md_request.cache_algo and md_request.push_to_cache:
            data_frame = self.speed_cache.get_dataframe(key)

        if data_frame is not None:
            return data_frame

        if md_request.split_request_chunks > 0:
            md_request_list = []

            if md_request.vendor_tickers is not None:

                tickers = copy.copy(md_request.tickers)
                vendor_tickers = copy.copy(md_request.vendor_tickers)

                for t, v in zip(tickers, vendor_tickers):
                    md = MarketDataRequest(md_request=md_request)
                    md.tickers = t
                    md.vendor_tickers = v
                    md.split_request_chunks = 0

                    md_request_list.append(md)

            return self.fetch_market(md_request_list)

        # Special cases when a predefined category has been asked
        if md_request.category is not None:

            if (md_request.category == 'fx-spot-volume' and md_request.data_source == 'quandl'):
                # NOT CURRENTLY IMPLEMENTED FOR FUTURE USE
                from findatapy.market.fxclsvolume import FXCLSVolume
                fxcls = FXCLSVolume(market_data_generator=self._market_data_generator)

                data_frame = fxcls.get_fx_volume(md_request.start_date, md_request.finish_date, md_request.tickers,
                                                 cut="LOC", data_source="quandl",
                                                 cache_algo=md_request.cache_algo)

            # For FX we have special methods for returning cross rates or total returns
            if (md_request.category in ['fx', 'fx-tot', 'fx-tot-forwards']) and md_request.tickers is not None and \
                    md_request.abstract_curve is None:
                fxcf = FXCrossFactory(market_data_generator=self._market_data_generator)

                if md_request.category == 'fx':
                    type = 'spot'
                elif md_request.category == 'fx-tot':
                    type = 'tot'

                elif md_request.category == 'fx-tot-forwards':
                    type = 'tot-forwards'

                if (md_request.freq != 'tick' and (md_request.fields == ['close'] or md_request.fields == ['open'])) or \
                        (md_request.freq == 'tick' and md_request.data_source in ['dukascopy', 'fxcm']):
                    data_frame = fxcf.get_fx_cross(md_request.start_date, md_request.finish_date, md_request.tickers,
                                                   cut=md_request.cut, data_source=md_request.data_source,
                                                   freq=md_request.freq,
                                                   cache_algo=md_request.cache_algo, type=type,
                                                   environment=md_request.environment, fields=md_request.fields,
                                                   data_engine=md_request.data_engine)

            # For FX implied volatility we can return the full surface
            if (md_request.category == 'fx-implied-vol'):
                if md_request.tickers is not None and md_request.freq == 'daily':
                    df = []

                    fxvf = FXVolFactory(market_data_generator=self._market_data_generator)

                    for t in md_request.tickers:
                        if len(t) == 6:
                            df.append(
                                fxvf.get_fx_implied_vol(md_request.start_date, md_request.finish_date, t, md_request.fx_vol_tenor,
                                                        cut=md_request.cut, data_source=md_request.data_source,
                                                        part=md_request.fx_vol_part,
                                                        cache_algo=md_request.cache_algo,
                                                        environment=md_request.environment,
                                                        field=md_request.fields,
                                                        data_engine=md_request.data_engine))

                    if df != []:
                        data_frame = Calculations().join(df, how='outer')

            # For FX vol market return all the market data necessary for pricing options
            # which includes FX spot, volatility surface, forward points, deposit rates
            if (md_request.category == 'fx-vol-market'):
                if md_request.tickers is not None:
                    df = []

                    fxcf = FXCrossFactory(market_data_generator=self._market_data_generator)
                    fxvf = FXVolFactory(market_data_generator=self._market_data_generator)
                    rates = RatesFactory(market_data_generator=self._market_data_generator)

                    # For each FX cross fetch the spot, vol and forward points
                    for t in md_request.tickers:
                        if len(t) == 6:
                            # Spot
                            df.append(
                                fxcf.get_fx_cross(start=md_request.start_date, end=md_request.finish_date, cross=t,
                                                  cut=md_request.cut, data_source=md_request.data_source,
                                                  freq=md_request.freq,
                                                  cache_algo=md_request.cache_algo, type='spot',
                                                  environment=md_request.environment,
                                                  fields=md_request.fields,
                                                  data_engine=md_request.data_engine))

                            # Entire FX vol surface
                            df.append(
                                fxvf.get_fx_implied_vol(md_request.start_date, md_request.finish_date, t, md_request.fx_vol_tenor,
                                                        cut=md_request.cut, data_source=md_request.data_source,
                                                        part=md_request.fx_vol_part,
                                                        cache_algo=md_request.cache_algo,
                                                        environment=md_request.environment,
                                                        field=md_request.fields,
                                                        data_engine=md_request.data_engine))

                            # FX forward points for every point on curve
                            df.append(rates.get_fx_forward_points(md_request.start_date, md_request.finish_date, t,
                                                                  md_request.fx_forwards_tenor,
                                                                  cut=md_request.cut,
                                                                  data_source=md_request.data_source,
                                                                  environment=md_request.environment,
                                                                  cache_algo=md_request.cache_algo,
                                                                  field=md_request.fields,
                                                                  data_engine=md_request.data_engine))

                    # Lastly fetch the base depos
                    df.append(rates.get_base_depos(md_request.start_date, md_request.finish_date,
                                                   self._get_base_depo_currencies(md_request.tickers), md_request.base_depos_tenor,
                                                   environment=md_request.environment,
                                                   cut=md_request.cut, data_source=md_request.data_source,
                                                   cache_algo=md_request.cache_algo, field=md_request.fields, data_engine=md_request.data_engine
                                                   ))

                    if df != []:
                        data_frame = Calculations().join(df, how='outer')

            if (md_request.category == 'fx-forwards-market'):
                if md_request.tickers is not None:
                    df = []

                    fxcf = FXCrossFactory(market_data_generator=self._market_data_generator)
                    rates = RatesFactory(market_data_generator=self._market_data_generator)

                    # For each FX cross fetch the spot and forward points
                    for t in md_request.tickers:
                        if len(t) == 6:
                            # Spot
                            df.append(
                                fxcf.get_fx_cross(start=md_request.start_date, end=md_request.finish_date, cross=t,
                                                  cut=md_request.cut, data_source=md_request.data_source,
                                                  freq=md_request.freq,
                                                  cache_algo=md_request.cache_algo, type='spot',
                                                  environment=md_request.environment,
                                                  fields=md_request.fields,
                                                  data_engine=md_request.data_engine))

                            # FX forward points for every point on curve
                            df.append(rates.get_fx_forward_points(md_request.start_date, md_request.finish_date, t,
                                                                  md_request.fx_forwards_tenor,
                                                                  cut=md_request.cut,
                                                                  data_source=md_request.data_source,
                                                                  environment=md_request.environment,
                                                                  cache_algo=md_request.cache_algo,
                                                                  field=md_request.fields,
                                                                  data_engine=md_request.data_engine))

                    # Lastly fetch the base depos
                    df.append(rates.get_base_depos(md_request.start_date, md_request.finish_date,
                                                   self._get_base_depo_currencies(md_request.tickers), md_request.base_depos_tenor,
                                                   cut=md_request.cut, data_source=md_request.data_source,
                                                   cache_algo=md_request.cache_algo,
                                                   environment=md_request.environment,
                                                   field=md_request.fields,
                                                   data_engine=md_request.data_engine
                                                   ))

                    if df != []:
                        data_frame = Calculations().join(df, how='outer')

            # eg. for calculating total return indices from first principles (rather than downloading them
            # from a data vendor
            if md_request.abstract_curve is not None:
                data_frame = md_request.abstract_curve.fetch_continuous_time_series \
                    (md_request, self._market_data_generator)

            if (md_request.category == 'crypto'):
                # Add more features later
                data_frame = self._market_data_generator.fetch_market_data(md_request)

            # TODO add more special examples here for different asset classes
            # the idea is that we do all the market data downloading here, rather than elsewhere

        # By default: pass the market data request to MarketDataGenerator
        # if data_frame is not None:
        #    data_frame = None

        if data_frame is None:
            data_frame = self._market_data_generator.fetch_market_data(md_request)

        # Special case where we can sometimes have duplicated data times
        if md_request.freq == 'intraday' and md_request.cut == 'BSTP':
            data_frame = self._filter.remove_duplicate_indices(data_frame)

        # Push into cache
        if md_request.push_to_cache:
            if data_frame is not None:
                self.speed_cache.put_dataframe(key, data_frame)

        return data_frame

    def create_md_request_from_dataframe(self, md_request_df, md_request=None, start_date=None, finish_date=None, smart_group=True,
                                          keep_initial_md_request_att_cols=['tickers', 'fields', 'freq'], **kwargs):

        md_list = []

        # Aggregate/shrink dataframe grouping it by common attributes
        # tickers, vendor_tickers, fields, vendor_fields
        if smart_group:
            md_request_df = ConfigManager().get_instance().smart_group_dataframe_tickers(
                md_request_df, ret_fields=md_request_df.columns.tolist())

        # Now populate MarketDataRequests based on the DataFrame
        for index, row in md_request_df.iterrows():
            if md_request is None:
                md_request_copy = MarketDataRequest()
            else:
                md_request_copy = MarketDataRequest(md_request=md_request)

            for col, val in row.items():

                try:
                    if ',' in val:
                        val = val.split(',')
                except:
                    pass

                # Only override the initial setting in md_request if they are None, or the user has specified this
                if getattr(md_request_copy, col) is None or col not in keep_initial_md_request_att_cols:
                    try:
                        if isinstance(val, list):
                            val = self.flatten_list_of_lists(val)

                        getattr(type(md_request_copy), col).fset(md_request_copy, val)
                    except:
                        pass

            if start_date is not None: md_request_copy.start_date = start_date
            if finish_date is not None: md_request_copy.finish_date = finish_date

            md_request_copy = self._kwargs_to_md_request(kwargs, md_request_copy)

            md_list.append(md_request_copy)

        return md_list

    def create_md_request_from_dict(self, md_request_dict, md_request=None, start_date=None, finish_date=None, **kwargs):

        if md_request is None:
            md_request = MarketDataRequest()

        for k in md_request_dict.keys():
            getattr(type(md_request), k).fset(md_request, md_request_dict[k])

        if start_date is not None: md_request.start_date = start_date
        if finish_date is not None: md_request.finish_date = finish_date

        md_request = self._kwargs_to_md_request(kwargs, md_request)

        return md_request

    def create_md_request_from_tickers(self, tickers, md_request=None, start_date=None, finish_date=None,
                                       best_match_only=False, smart_group=True, **kwargs):
        md_request_list = []

        if isinstance(tickers, str):
            tickers = [tickers]

        for t in tickers:
            md_request_copy = self.create_md_request_from_str(
                '_.' + t, md_request=md_request, start_date=start_date, finish_date=finish_date, best_match_only=best_match_only, smart_group=smart_group)

            md_request_copy = self._kwargs_to_md_request(kwargs, md_request_copy)
            md_request_list.append(md_request_copy)

        return self.flatten_list_of_lists(md_request_list)

    def _kwargs_to_md_request(self, kw, md_request):

        if not(isinstance(md_request, list)):
            # Any kwargs are assumed to be to set MarketDataRequest attributes
            if kw != {}:

                if md_request is None:
                    md_request = MarketDataRequest()

                for k in kw.keys():
                    setattr(md_request, k, kw[k])

            return md_request

        # Any kwargs are assumed to be to set MarketDataRequest attributes
        if kw != {}:
            md_request_mod_list = []

            for md in md_request:
                if md is None:
                    md = MarketDataRequest()

                for k in kw.keys():
                    setattr(md, k, kw[k])

                md_request_mod_list.append(md)

            md_request = md_request_mod_list

        return md_request

    def create_md_request_from_str(self, md_request_str, md_request=None, start_date=None, finish_date=None, best_match_only=False,
                                   smart_group=True, **kwargs):

        json_md_request = None

        # Try to parse str as JSON if that fails, then try as a str
        try:
            json_md_request = json.loads(md_request_str)

            if md_request is None:
                md_request = MarketDataRequest()

            for k in json_md_request.keys():
                getattr(type(md_request), k).fset(md_request, json_md_request[k])

            if start_date is not None: md_request.start_date = start_date
            if finish_date is not None: md_request.finish_date = finish_date

            return md_request
        except:
            pass

        # If we failed to parse as a JSON, let's try as string
        if json_md_request is None:

            # Split up ticker string
            md_request_params = ConfigManager.split_ticker_string(md_request_str)

            environment = md_request_params[0]

            # The user can omit the environment
            if environment in constants.possible_data_environment:
                i = 0

            # Otherwise, what if the user wants to specify each property manually?
            elif environment == 'raw' or environment == 'r':
                # Here the user can specify any tickers/fields etc. they want, they don't have to be predefined
                # eg. raw.data_source.bloomberg.tickers.EURUSD.vendor_tickers.EURUSD Curncy
                if md_request is None:
                    md_request = MarketDataRequest()

                freeform_md_request = {}

                for c in range(1, len(md_request_params), 2):
                    f = md_request_params[c + 1]

                    if ',' in f:
                        f = f.split(',')

                    freeform_md_request[md_request_params[c]] = f

                md_request.freeform_md_request = freeform_md_request

                if start_date is not None: md_request.start_date = start_date
                if finish_date is not None: md_request.finish_date = finish_date

                md_request = self.create_md_request_from_freeform(md_request)
                md_request = self._kwargs_to_md_request(kwargs, md_request)

                return md_request

            # Otherwise we do a partial match of predefined tickers
            elif environment == "_":
                # Try a heuristic/approximate match eg. _.quandl.fx
                md_request_df = ConfigManager().get_instance().free_form_tickers_query(md_request_params[1:],
                                                                                       best_match_only=best_match_only,
                                                                                       smart_group=smart_group)

                md_request_df = self.create_md_request_from_dataframe(md_request_df,
                                                             md_request=md_request, start_date=start_date,
                                                             finish_date=finish_date)

                md_request_df = self._kwargs_to_md_request(kwargs, md_request_df)

                # if best_match_only:
                return md_request_df


            else:
                i = -1
                environment = None

            # Otherwise the user has specified the MarketDataRequest str in the form
            # category.data_source.freq.cut.tickers.field = fx.bloomberg.daily.NYC.EURUSD.close
            category = md_request_params[i + 1]
            data_source = md_request_params[i + 2]

            #for f in fileformat_reverse.keys():
            #    md_request_str = md_request_str.replace(f, fileformat_reverse[f])

            cut = None
            freq = None
            tickers = None
            fields = None

            # The freq, cut, tickers, fields are optional (in which case defaults will be used)
            try:
                freq = md_request_params[i + 3]
            except:
                pass

            try:
                cut = md_request_params[i + 4]
            except:
                pass

            # We can have multiple tickers and fields separately by a ticker
            try:
                tickers = md_request_params[i + 5]

                if ',' in tickers:
                    tickers = tickers.split(',')
            except:
                pass

            try:
                fields = md_request_params[i + 6]

                if ',' in fields:
                    fields = fields.split(',')
            except:
                pass

            if md_request is None:
                md_request = MarketDataRequest(category=category, data_source=data_source)
            else:
                md_request.category = category
                md_request.data_source = data_source

            if environment is not None: md_request.environment = environment
            if start_date is not None: md_request.start_date = start_date
            if finish_date is not None: md_request.finish_date = finish_date
            if freq is not None: md_request.freq = freq
            if cut is not None: md_request.cut = cut
            if tickers is not None: md_request.tickers = tickers
            if fields is not None: md_request.fields = fields

        return md_request

    def create_md_request_from_freeform(self, md_request, freeform_md_request=None, return_df=False, **kwargs):

        if freeform_md_request is None:
            freeform_md_request = md_request.freeform_md_request

        if freeform_md_request is None:
            return md_request

        if isinstance(freeform_md_request, pd.DataFrame):
            pass
        else:
            if not(isinstance(freeform_md_request, list)):
                freeform_md_request = [freeform_md_request]

            df = pd.DataFrame.from_dict(freeform_md_request)

        group_columns = [x for x in df.columns if x not in ['tickers', 'vendor_tickers']]

        # Group by everything but the tickers, vendor_tickers (which are concatenated)
        # Make sure we remove any duplicated tickers
        if group_columns == []:
            if 'tickers' in df.columns and 'vendor_tickers' not in df.columns:
                df['tickers'] = self.remove_duplicates_and_flatten_list(df['tickers'].values.tolist())
                freeform_md_request = [{'tickers' : df['tickers'].values.tolist()}]
            elif 'tickers' in df.columns and 'vendor_tickers' in df.columns:
                df['tickers'] = self.remove_duplicates_and_flatten_list(df['tickers'].values.tolist())
                df['vendor_tickers'] = self.remove_duplicates_and_flatten_list(df['vendor_tickers'].values.tolist())
                freeform_md_request = [{'tickers': df['tickers'].values.tolist(), 'vendor_tickers' : df['vendor_tickers'].values.tolist()}]
        else:
            if 'tickers' in df.columns and 'vendor_tickers' not in df.columns:
                df = df.groupby(group_columns, as_index=False).agg({'tickers': list})

                for i in range(0, len(df.index)):
                    df.at[i, 'tickers'] = self.remove_duplicates_and_flatten_list(df.iloc[i]['tickers'])

            elif 'tickers' in df.columns and 'vendor_tickers' in df.columns:
                df = df.groupby(group_columns, as_index=False).agg({'tickers': list, 'vendor_tickers' : list})

                for i in range(0, len(df.index)):
                    df.at[i, 'tickers'] = self.remove_duplicates_and_flatten_list(df.iloc[i]['tickers'])
                    df.at[i, 'vendor_tickers'] = self.remove_duplicates_and_flatten_list(df.iloc[i]['vendor_tickers'])

            freeform_md_request = df.to_dict('records')

        md_request_copy = MarketDataRequest(md_request=md_request)
        md_request_copy.freeform_md_request = None

        md_request_list = []

        for f in freeform_md_request:
            md_request_temp = MarketDataRequest(md_request=md_request_copy)

            for k in f.keys():
                lst = f[k]

                if isinstance(f[k], list):

                    # Make sure we don't have duplicated tickers requested (or indeed any other values)
                    lst = self.remove_duplicates_and_flatten_list(f[k])

                getattr(type(md_request_temp), k).fset(md_request_temp, lst)

            md_request_temp = self._kwargs_to_md_request(kwargs, md_request_temp)

            md_request_list.append(md_request_temp)

        if len(md_request_list) == 1:
            md_request_list = md_request_list[0]

        if return_df:
            return md_request_list, df

        return md_request_list

    def remove_list_duplicates(self, lst):
        return list(dict.fromkeys(lst))

    def remove_duplicates_and_flatten_list(self, lst):
        return list(dict.fromkeys(self.flatten_list_of_lists(lst)))

    def flatten_list_of_lists(self, list_of_lists):
        """Flattens lists of obj, into a single list of strings (rather than characters, which is default behavior).

        Parameters
        ----------
        list_of_lists : obj (list)
            List to be flattened

        Returns
        -------
        str (list)
        """

        if isinstance(list_of_lists, list):
            rt = []
            for i in list_of_lists:
                if isinstance(i, list):
                    rt.extend(self.flatten_list_of_lists(i))
                else:
                    rt.append(i)

            return rt

        return list_of_lists

    def _get_base_depo_currencies(self, cross):

        if not(isinstance(cross, list)):
            cross = [cross]

        base_depo_currencies = []

        for c in cross:
            base = c[0:3]; terms = c[3:6]

            if base in constants.base_depos_currencies:
                base_depo_currencies.append(base)

            if terms in constants.base_depos_currencies:
                base_depo_currencies.append(terms)

        base_depo_currencies = list(set(base_depo_currencies))

        return base_depo_currencies


########################################################################################################################

from findatapy.util.fxconv import FXConv


class FXCrossFactory(object):
    """Generates FX spot time series and FX total return time series (assuming we already have
    total return indices available from xxxUSD form) from underlying series. Can also produce cross rates from the USD
    crosses.

    """

    def __init__(self, market_data_generator=None):
        self.fxconv = FXConv()

        self.cache = {}

        self._calculations = Calculations()
        self._market_data_generator = market_data_generator

        return

    def get_fx_cross_tick(self, start, end, cross,
                          cut="NYC", data_source="dukascopy", cache_algo='internet_load_return', type='spot',
                          environment=constants.default_data_environment, fields=['bid', 'ask'], data_engine=constants.default_data_engine):

        if isinstance(cross, str):
            cross = [cross]

        market_data_request = MarketDataRequest(
            gran_freq="tick",
            freq_mult=1,
            freq='tick',
            cut=cut,
            fields=['bid', 'ask', 'bidv', 'askv'],
            cache_algo=cache_algo,
            environment=environment,
            start_date=start,
            finish_date=end,
            data_source=data_source,
            category='fx',
            data_engine=data_engine
        )

        market_data_generator = self._market_data_generator
        data_frame_agg = None

        for cr in cross:

            if (type == 'spot'):
                market_data_request.tickers = cr

                cross_vals = market_data_generator.fetch_market_data(market_data_request)

                if cross_vals is not None:

                    # If user only wants 'close' calculate that from the bid/ask fields
                    if fields == ['close']:
                        cross_vals = pd.DataFrame(cross_vals[[cr + '.bid', cr + '.ask']].mean(axis=1))
                        cross_vals.columns = [cr + '.close']
                    else:
                        filter = Filter()

                        filter_columns = [cr + '.' + f for f in fields]
                        cross_vals = filter.filter_time_series_by_columns(filter_columns, cross_vals)

            if data_frame_agg is None:
                data_frame_agg = cross_vals
            else:
                data_frame_agg = data_frame_agg.join(cross_vals, how='outer')

        if data_frame_agg is not None:
            # Strip the nan elements
            data_frame_agg = data_frame_agg.dropna()

        return data_frame_agg

    def get_fx_cross(self, start, end, cross,
                     cut="NYC", data_source="bloomberg", freq="intraday", cache_algo='internet_load_return',
                     type='spot',
                     environment=constants.default_data_environment, fields=['close'], data_engine=constants.default_data_engine):

        if data_source == "gain" or data_source == 'dukascopy' or freq == 'tick':
            return self.get_fx_cross_tick(start, end, cross,
                                          cut=cut, data_source=data_source, cache_algo=cache_algo, type='spot',
                                          environment=environment,
                                          fields=fields, data_engine=data_engine)

        if isinstance(cross, str):
            cross = [cross]

        market_data_request_list = []
        freq_list = []
        type_list = []

        for cr in cross:
            market_data_request = MarketDataRequest(freq_mult=1,
                                                    cut=cut,
                                                    fields=fields,
                                                    freq=freq,
                                                    cache_algo=cache_algo,
                                                    start_date=start,
                                                    finish_date=end,
                                                    data_source=data_source,
                                                    environment=environment,
                                                    data_engine=data_engine)

            market_data_request.type = type
            market_data_request.cross = cr

            if freq == 'intraday':
                market_data_request.gran_freq = "minute" # intraday

            elif freq == 'daily':
                market_data_request.gran_freq = "daily" # daily

            market_data_request_list.append(market_data_request)

        data_frame_agg = []

        # Depends on the nature of operation as to whether we should use threading or multiprocessing library
        if constants.market_thread_technique == "thread":
            from multiprocessing.dummy import Pool
        else:
            # Most of the time is spend waiting for Bloomberg to return, so can use threads rather than multiprocessing
            # must use the multiprocess library otherwise can't pickle objects correctly
            # note: currently not very stable
            from multiprocess import Pool

        thread_no = constants.market_thread_no['other']

        if market_data_request_list[0].data_source in constants.market_thread_no:
            thread_no = constants.market_thread_no[market_data_request_list[0].data_source]

        # Fudge, issue with multithreading and accessing HDF5 files
        # if self._market_data_generator.__class__.__name__ == 'CachedMarketDataGenerator':
        #    thread_no = 0
        thread_no = 0

        if (thread_no > 0):
            pool = Pool(thread_no)

            # Open the market data downloads in their own threads and return the results
            df_list = pool.map_async(self._get_individual_fx_cross, market_data_request_list).get()

            data_frame_agg = self._calculations.join(df_list, how='outer')

            # data_frame_agg = self._calculations.pandas_outer_join(result.get())

            try:
                pool.close()
                pool.join()
            except:
                pass
        else:
            for md_request in market_data_request_list:
                data_frame_agg.append(self._get_individual_fx_cross(md_request))

            data_frame_agg = self._calculations.join(data_frame_agg, how='outer')

        # Strip the nan elements
        data_frame_agg = data_frame_agg.dropna(how='all')

        # self.speed_cache.put_dataframe(key, data_frame_agg)

        return data_frame_agg

    def _get_individual_fx_cross(self, market_data_request):
        cr = market_data_request.cross
        type = market_data_request.type
        freq = market_data_request.freq

        base = cr[0:3]
        terms = cr[3:6]

        if (type == 'spot'):
            # Non-USD crosses
            if base != 'USD' and terms != 'USD':
                base_USD = self.fxconv.correct_notation('USD' + base)
                terms_USD = self.fxconv.correct_notation('USD' + terms)

                # TODO check if the cross exists in the database

                # Download base USD cross
                market_data_request.tickers = base_USD
                market_data_request.category = 'fx'

                base_vals = self._market_data_generator.fetch_market_data(market_data_request)

                # Download terms USD cross
                market_data_request.tickers = terms_USD
                market_data_request.category = 'fx'

                terms_vals = self._market_data_generator.fetch_market_data(market_data_request)

                # If quoted USD/base flip to get USD terms
                if (base_USD[0:3] == 'USD'):
                    base_vals = 1 / base_vals

                # If quoted USD/terms flip to get USD terms
                if (terms_USD[0:3] == 'USD'):
                    terms_vals = 1 / terms_vals

                base_vals.columns = ['temp'];
                terms_vals.columns = ['temp']

                cross_vals = base_vals.div(terms_vals, axis='index')
                cross_vals.columns = [cr + '.' +  market_data_request.fields[0]]

                base_vals.columns = [base_USD + '.' + market_data_request.fields[0]]
                terms_vals.columns = [terms_USD + '.' + market_data_request.fields[0]]
            else:
                # if base == 'USD': non_USD = terms
                # if terms == 'USD': non_USD = base

                correct_cr = self.fxconv.correct_notation(cr)

                market_data_request.tickers = correct_cr
                market_data_request.category = 'fx'

                cross_vals = self._market_data_generator.fetch_market_data(market_data_request)

                # Special case for USDUSD!
                if base + terms == 'USDUSD':
                    if freq == 'daily':
                        cross_vals = pd.DataFrame(1, index=cross_vals.index, columns=cross_vals.columns)
                        filter = Filter()
                        cross_vals = filter.filter_time_series_by_holidays(cross_vals, cal='WEEKDAY')
                else:
                    # Flip if not convention (eg. JPYUSD)
                    if (correct_cr != cr):
                        cross_vals = 1 / cross_vals

                # cross_vals = self._market_data_generator.harvest_time_series(md_request)
                cross_vals.columns = [cr + '.' + market_data_request.fields[0]]

        elif type[0:3] == "tot":
            if freq == 'daily':
                # Download base USD cross
                market_data_request.tickers = base + 'USD'
                market_data_request.category = 'fx-' + type

                if type[0:3] == "tot":
                    base_vals = self._market_data_generator.fetch_market_data(market_data_request)

                # Download terms USD cross
                market_data_request.tickers = terms + 'USD'
                market_data_request.category = 'fx-' + type

                if type[0:3] == "tot":
                    terms_vals = self._market_data_generator.fetch_market_data(market_data_request)

                # base_rets = self._calculations.calculate_returns(base_vals)
                # terms_rets = self._calculations.calculate_returns(terms_vals)

                # Special case for USDUSD case (and if base or terms USD are USDUSD
                if base + terms == 'USDUSD':
                    base_rets = self._calculations.calculate_returns(base_vals)
                    cross_rets = pd.DataFrame(0, index=base_rets.index, columns=base_rets.columns)
                elif base + 'USD' == 'USDUSD':
                    cross_rets = -self._calculations.calculate_returns(terms_vals)
                elif terms + 'USD' == 'USDUSD':
                    cross_rets = self._calculations.calculate_returns(base_vals)
                else:
                    base_rets = self._calculations.calculate_returns(base_vals)
                    terms_rets = self._calculations.calculate_returns(terms_vals)

                    cross_rets = base_rets.sub(terms_rets.iloc[:, 0], axis=0)

                # First returns of a time series will by NaN, given we don't know previous point
                cross_rets.iloc[0] = 0

                cross_vals = self._calculations.create_mult_index(cross_rets)
                cross_vals.columns = [cr + '-' + type + '.' +  market_data_request.fields[0]]

            elif freq == 'intraday':
                LoggerManager().getLogger(__name__).info('Total calculated returns for intraday not implemented yet')
                return None

        return cross_vals

#######################################################################################################################

import pandas as pd

from findatapy.market.marketdatarequest import MarketDataRequest
from findatapy.util import LoggerManager
from findatapy.timeseries import Calculations, Filter, Timezone

class FXVolFactory(object):
    """Generates FX implied volatility time series and surfaces (using very simple interpolation!) and only in delta space.

    """

    def __init__(self, market_data_generator=None):

        self._market_data_generator = market_data_generator

        self._calculations = Calculations()
        self._filter = Filter()
        self._timezone = Timezone()

        self._rates = RatesFactory()

        return

    def get_fx_implied_vol(self, start, end, cross, tenor, cut="BGN", data_source="bloomberg", part="V",
                           cache_algo="internet_load_return", environment=constants.default_data_environment, field='close',
                           data_engine=constants.default_data_engine):
        """Get implied vol for specified cross, tenor and part of surface. By default we use Bloomberg, but we could
        use any data provider for which we have vol tickers.

        Note, that for Bloomberg not every point will be quoted for each dataset (typically, BGN will have more points
        than for example LDN)

        Parameters
        ----------
        start : datetime
            start date of request
        end : datetime
            end date of request
        cross : str
            FX cross
        tenor : str
            tenor of implied vol
        cut : str
            closing time of data
        data_source : str
            data_source of market data eg. bloomberg
        part : str
            part of vol surface eg. V for ATM implied vol, 25R 25 delta risk reversal

        Return
        ------
        pd.DataFrame
        """

        market_data_generator = self._market_data_generator

        if tenor is None:
            tenor = constants.fx_vol_tenor

        if part is None:
            part = constants.fx_vol_part

        tickers = self.get_labels(cross, part, tenor)

        market_data_request = MarketDataRequest(
            start_date=start, finish_date=end,
            data_source=data_source,
            category='fx-implied-vol',
            freq='daily',
            cut=cut,
            tickers=tickers,
            fields=field,
            cache_algo=cache_algo,
            environment=environment,
            data_engine=data_engine
        )

        data_frame = market_data_generator.fetch_market_data(market_data_request)
        # data_frame.index.name = 'Date'

        # Special case for 10AM NYC cut
        # - get some historical 10AM NYC data (only available on BBG for a few years, before 2007)
        # - fill the rest with a weighted average of TOK/LDN closes
        if cut == "10AM":
            # Where we have actual 10am NY data use that & overwrite earlier estimated data (next)
            vol_data_10am = data_frame

            # As for most dates we probably won't have 10am data, so drop rows where there's no data at all
            # Can have the situation where some data won't be there (eg. longer dated illiquid tenors)
            if vol_data_10am is not None:
                vol_data_10am = vol_data_10am.dropna(how='all')  # Only have limited ON 10am cut data

            # Now get LDN and TOK vol data to fill any gaps
            vol_data_LDN = self.get_fx_implied_vol(start=start, end=end, cross=cross, tenor=tenor,
                                                                       data_source=data_source, cut='LDN', part=part,
                                                                       cache_algo=cache_algo, field=field, data_engine=data_engine)

            vol_data_TOK = self.get_fx_implied_vol(start=start, end=end, cross=cross, tenor=tenor,
                                                                       data_source=data_source, cut='TOK', part=part,
                                                                       cache_algo=cache_algo, field=field, data_engine=data_engine)

            # vol_data_LDN.index = pandas.DatetimeIndex(vol_data_LDN.index)
            # vol_data_TOK.index = pandas.DatetimeIndex(vol_data_TOK.index)

            old_cols = vol_data_LDN.columns

            vol_data_LDN.columns = vol_data_LDN.columns.values + "LDN"
            vol_data_TOK.columns = vol_data_TOK.columns.values + "TOK"

            data_frame = vol_data_LDN.join(vol_data_TOK, how='outer')

            # Create very naive average of LDN and TOK to estimate 10am NY value because we often don't have this data
            # Note, this isn't perfect, particularly on days where you have payrolls data, and we're looking at ON data
            # You might choose to create your own approximation for 10am NY
            for col in old_cols:
                data_frame[col] = (1 * data_frame[col + "LDN"] + 3 * data_frame[col + "TOK"]) / 4
                # data_frame[col] = data_frame[col + "LDN"]
                data_frame.pop(col + "LDN")
                data_frame.pop(col + "TOK")

            # Get TOK/LDN vol data before 10am and after 10am (10am data is only available for a few years)
            # If we have no original 10am data don't bother
            if vol_data_10am is not None:
                if not(vol_data_10am.empty):
                    pre_vol_data = data_frame[data_frame.index < vol_data_10am.index[0]]
                    post_vol_data = data_frame[data_frame.index > vol_data_10am.index[-1]]

                    data_frame = (pre_vol_data.append(vol_data_10am)).append(post_vol_data)

            # data_frame.index = pandas.to_datetime(data_frame.index)

        return data_frame

    def get_labels(self, cross, part, tenor):
        if isinstance(cross, str): cross = [cross]
        if isinstance(tenor, str): tenor = [tenor]
        if isinstance(part, str): part = [part]

        tickers = []

        for cr in cross:
            for tn in tenor:
                for pt in part:
                    tickers.append(cr + pt + tn)

        return tickers

    def extract_vol_surface_for_date(self, df, cross, date_index, delta=constants.fx_vol_delta, tenor=constants.fx_vol_tenor, field='close'):
        """Get's the vol surface in delta space without any interpolation

        Parameters
        ----------
        df : DataFrame
            With vol data
        cross : str
            Currency pair
        date_index : int
            Which date to extract
        delta : list(int)
            Deltas which are quoted in order of out-of-money -> in-the-money (eg. [10, 25])
        tenor : list(str)
            Tenors which are quoted (eg. ["ON", "1W"...]

        Returns
        -------
        DataFrame
        """

        # Assume we have a matrix of the form
        # eg. EURUSDVON.close ...

        # types of quotation on vol surface
        # self.part = ["V", "25R", "10R", "25B", "10B"]

        # all the tenors on our vol surface
        # self.tenor = ["ON", "1W", "2W", "3W", "1M", "2M", "3M", "6M", "9M", "1Y", "2Y", "5Y"]

        strikes = []

        for d in delta:
            strikes.append(str(d) + "DP")

        strikes.append('ATM')

        for d in delta:
            strikes.append(str(d) + "DC")

        df_surf = pd.DataFrame(index=strikes, columns=tenor)

        for ten in tenor:
            for d in delta:
                df_surf[ten][str(d) + "DP"] = df[cross + "V" + ten + "."+ field][date_index] \
                                          - (df[cross + str(d) +"R" + ten + "." + field][date_index] / 2.0) \
                                          + (df[cross + str(d) + "B" + ten + "." + field][date_index])

                df_surf[ten][str(d) + "DC"] = df[cross + "V" + ten + "." + field][date_index] \
                                          + (df[cross + str(d) + "R" + ten + "." + field][date_index] / 2.0) \
                                          + (df[cross + str(d) + "B" + ten + "." + field][date_index])

            df_surf[ten]["ATM"] = df[cross + "V" + ten + "." + field][date_index]

        return df_surf


#######################################################################################################################

class RatesFactory(object):
    """Gets the deposit rates for a particular currency (or forwards for a currency pair)

    """

    def __init__(self, market_data_generator=None):

        self.cache = {}

        self._calculations = Calculations()
        self._market_data_generator = market_data_generator

        return

    def get_base_depos(self, start, end, currencies, tenor, cut="NYC", data_source="bloomberg",
                       cache_algo="internet_load_return", field='close', environment=constants.default_data_environment,
                       data_engine=constants.default_data_engine):
        """Gets the deposit rates for a particular tenor and part of surface

        Parameter
        ---------
        start : DateTime
            Start date
        end : DateTime
            End data
        currencies : str
            Currencies for which we want to download deposit rates
        tenor : str
            Tenor of deposit rate
        cut : str
            Closing time of the market data
        data_source : str
            data_source of the market data eg. bloomberg
        cache_algo : str
            Caching scheme for the data

        Returns
        -------
        pd.DataFrame
            Contains deposit rates
        """

        market_data_generator = self._market_data_generator

        if tenor is None:
            tenor = constants.base_depos_tenor

        if isinstance(currencies, str): currencies = [currencies]
        if isinstance(tenor, str): tenor = [tenor]

        tickers = []

        for cr in currencies:

            for tn in tenor:
                tickers.append(cr + tn)

        # Special case for Fed Funds Effective Rate which we add in all instances
        if 'USDFedEffectiveRate' not in tickers:
            tickers.append("USDFedEffectiveRate")

        # For depos there usually isn't a 10AM NYC cut available, so just use TOK data
        # Also no BGN tends to available for deposits, so use NYC
        if cut == '10AM':
            cut = 'TOK'
        elif cut == 'BGN':
            cut = 'NYC'

        market_data_request = MarketDataRequest(
            start_date=start, finish_date=end,
            data_source=data_source,
            category='base-depos',
            freq='daily',
            cut=cut,
            tickers=tickers,
            fields=field,
            cache_algo=cache_algo,
            environment=environment,
            data_engine=data_engine
        )

        data_frame = market_data_generator.fetch_market_data(market_data_request)
        data_frame.index.name = 'Date'

        return data_frame

    def get_fx_forward_points(self, start, end, cross, tenor, cut="BGN", data_source="bloomberg",
                              cache_algo="internet_load_return", field='close', environment=constants.default_data_environment,
                              data_engine=constants.default_data_engine):
        """Gets the forward points for a particular tenor and currency

        Parameter
        ---------
        start : Datetime
            Start date
        end : Datetime
            End data
        cross : str
            FX crosses for which we want to download forward points
        tenor : str
            Tenor of deposit rate
        cut : str
            Closing time of the market data
        data_source : str
            data_source of the market data eg. bloomberg
        cache_algo : str
            Caching scheme for the data

        Returns
        -------
        pd.DataFrame
        Contains deposit rates
        """

        # md_request = MarketDataRequest()
        market_data_generator = self._market_data_generator

        # md_request.data_source = data_source  # use bbg as a data_source
        # md_request.start_date = start  # start_date
        # md_request.finish_date = end  # finish_date

        if tenor is None:
            tenor = constants.fx_forwards_tenor

        if isinstance(cross, str): cross = [cross]
        if isinstance(tenor, str): tenor = [tenor]

        # Tickers are often different on Bloomberg for forwards/depos vs vol, so want consistency so 12M is always 1Y
        tenor = [x.replace('1Y', '12M') for x in tenor]

        tickers = []

        for cr in cross:
            for tn in tenor:
                tickers.append(cr + tn)

        market_data_request = MarketDataRequest(
            start_date=start, finish_date=end,
            data_source=data_source,
            category='fx-forwards',
            freq='daily',
            cut=cut,
            tickers=tickers,
            fields=field,
            cache_algo=cache_algo,
            environment=environment,
            data_engine=data_engine
        )

        data_frame = market_data_generator.fetch_market_data(market_data_request)
        data_frame.columns = [x.replace('12M', '1Y') for x in data_frame.columns]
        data_frame.index.name = 'Date'

        return data_frame



__author__ = 'saeedamen'

from findatapy.market.datavendor import DataVendor

# don't include DataVendorBBG, in case users haven't installed blpapi
# from findatapy.market.datavendorbbg import DataVendorBBG
from findatapy.market.ioengine import IOEngine, SpeedCache
from findatapy.market.market import Market, FXVolFactory, FXCrossFactory, FXConv, RatesFactory
from findatapy.market.marketdatagenerator import MarketDataGenerator
from findatapy.market.marketdatarequest import MarketDataRequest



__author__ = 'saeedamen' # Saeed Amen

#
# Copyright 2016 Cuemacro
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and limitations under the License.
#

from findatapy.util import DataConstants

class SwimPool(object):
    """Creating thread and process pools in a generic way. Allows users to specify the underlying thread or multiprocess library
    they wish to use. Note you can share Pool objects between processes.

    """

    def __init__(self, multiprocessing_library = None):
        self._pool = None

        if multiprocessing_library is None:
            multiprocessing_library = DataConstants().multiprocessing_library

        self._multiprocessing_library = multiprocessing_library
        self._thread_technique = 'na'

        if multiprocessing_library == 'multiprocess':
            try:
                import multiprocess;
                multiprocess.freeze_support()
            except:
                pass
        elif multiprocessing_library == 'multiprocessing_on_dill':
            try:
                import multiprocessing_on_dill;
                multiprocessing_on_dill.freeze_support()
            except:
                pass
        elif multiprocessing_library == 'multiprocessing':
            try:
                import multiprocessing;
                multiprocessing.freeze_support()
            except:
                pass

    def create_pool(self, thread_technique, thread_no, force_new = True, run_in_parallel = True):

        self._thread_technique = thread_technique

        if not(force_new) and self._pool is not None:
            return self._pool

        if thread_technique == "thread" or run_in_parallel == False:
            from multiprocessing.dummy import Pool
        elif thread_technique == "multiprocessing":
            # most of the time is spend waiting for Bloomberg to return, so can use threads rather than multiprocessing
            # must use the multiprocessing_on_dill library otherwise can't pickle objects correctly
            # note: currently not very stable
            if self._multiprocessing_library == 'multiprocessing_on_dill':
                from multiprocessing_on_dill import Pool
            elif self._multiprocessing_library == 'multiprocess':
                from multiprocess import Pool
            elif self._multiprocessing_library == 'multiprocessing':
                from multiprocessing import Pool
            elif self._multiprocessing_library == 'pathos':
                from pathos.multiprocessing import Pool
                #from pathos.pools import ProcessPool as Pool
            elif self._multiprocessing_library == 'billiard':
                from billiard.pool import Pool

        if run_in_parallel == False: thread_no = 1

        self._pool = Pool(thread_no)

        return self._pool

    def close_pool(self, pool, force_process_respawn=False):
        if pool is not None:
            if (self._thread_technique != 'multiprocessing' and self._multiprocessing_library != 'pathos') \
                    or force_process_respawn:
                pool.close()
                pool.join()